#
# unimacroutils.py (in dtactions/unimacro)
# was module natlinkutilsqh.py in unimacro
#
#  Quintijn Hoogenboom
#  February 2000/.../March 2021/July 2021
#
#  
"""a set of utility functions in dtactions/unimacro

Previous this was natlinkutilsqh.py in unimacro

"""
#pylint:disable=C0302, C0116, C0321, R0912, R0913, R0914, R0915, W0613
#pylint:disable=E1101
import time
import re
import os
import os.path
import sys
import collections
from pathlib import Path
import win32gui
import win32clipboard
# import pywintypes

from dtactions import monitorfunctions  # elaborated version QH
from dtactions import autohotkeyactions

import natlink
from natlinkcore import natlinkutils
from natlinkcore import natlinkstatus
status = natlinkstatus.NatlinkStatus()

DEBUG = 0
# errors for (mainly) the command grammar:

class NatlinkCommandError(Exception):
    """NatlinkCommandError"""
class NatlinkCommandTimeOut(Exception):
    """NatlinkCommandTimeOut"""
class UnimacroError(Exception):
    """UnimacroError"""

# default Waiting Times:
defaultWaitingTime = 0.1        # 100 milliseconds
visibleWaitFactor = 3                 # default for Wait and W
longWaitFactor = 10                   # times 3 for visible wait
shortWaitFactor = 0.3                 # times 10 for long wait
                                # times 0.3 for short wait
#debugMode 0 = not, -1 == dvcMode, 1 is light, 2 = normal, 3 = heavy
debugMode = 1

pendingExecScripts = []

ProgInfo = collections.namedtuple('ProgInfo', 'progpath prog title toporchild classname hndle'.split(' '))



# class CaseInsensitiveSet(set):
#     """makes the comparison for adding new elements case insensitive
#     Adding takes the capitalisation of the first item added
#     """
#     def __hash__(self):
#         return hash(self.lower())
#     def __eq__(self, other):
#         return self.lower() == other.lower()

def getLanguage():
    """get current language, 3 characters

    """
    return status.get_language()

def getUserLanguage():
    """get current language, long name

    """
    raise NotImplementedError('getUserLanguage (the long language name) is no longer available, use natlinkstatus function "get_language"')

# utility functions----------------------------------------
## matchWindow from natlinkutils:
##def matchWindow(moduleInfo, modName, wndText):
##    if len(moduleInfo)<3 or not moduleInfo[0]: return None
##    curName = getBaseName(moduleInfo[0].lower() )
##    if curName != modName: return None
##    if -1 == moduleInfo[1].find(wndText): return None
##    return moduleInfo[2]

# A utility function which determines whether modInfo matches a specified
# module name.
# If modInfo is not given, it is got here.
# 
# Returns module name (program name!) on match and None on mismatch.
# This is a variant on matchWindow (from Joel), but now you can also check only the module, 
# and not the (sub) window of this module.
# Also you can specify a window title or a list/tuple of window titles.
# If you specify a window, it can be checked with exact title or exact case.
# If nothing is specified lower case strings are compared, and only part
# of the window title has to be given.
def getCurrentModuleSafe(nWait=5, waitingTime=0.01):
    """in case natlink.getCurrentModule returns None, try again a few times...
    default 5 x 0.01 = 0.05 (50 milliseconds)
    
    returns the tuple: 
    (full file name of foreground program, window title, window handle)
    """
    modInfo = natlink.getCurrentModule()
    if modInfo:
        return modInfo
    for i in range(nWait):
        time.sleep(waitingTime)
        modInfo = natlink.getCurrentModule()
        waited = (i+1)*waitingTime
        if modInfo:
            print(f'getCurrentModuleSafe, found modInfo after {waited} seconds')
            return modInfo

    print(f'getCurrentModuleSafe, could not find modInfo after {waited} seconds')
    return None

def matchModule(modName, wantedTitle = None, modInfo=None, titleExact=0, caseExact=0):
    """return modName if title matches, otherwise None
    """
    #pylint:disable=R0911
    if modInfo is None:
        modInfo = natlink.getCurrentModule()
    if not modInfo[0]:
        return None
    modName = modName.lower()
##    print 'modName: %s, basename module: %s'% (modName, getBaseName(modInfo[0]))
    if modName != getBaseNameLower(modInfo[0]):
        return None
    if not wantedTitle:
        return modName
   
    winTitle = modInfo[1]
    if isinstance(wantedTitle, str):    
        if not caseExact:
            wantedTitle = wantedTitle.lower()
            winTitle = winTitle.lower()
        if titleExact:
            if winTitle == wantedTitle:
                return modName
        else:
            if winTitle.find(wantedTitle) >= 0:
                return modName
    elif isinstance(wantedTitle, (list, tuple)):
        for t in wantedTitle:
            if matchModule(modName, t, modInfo, titleExact, caseExact):
                return modName
    return None

# Matches if only the window title ok. If modInfo is not given, it is extracted here.
#
# The wantedTitle can also be a list or tuple, in that case the routine is called
# recursive.
# titleExact and caseExact can be given.  If not strings are converted to lower case
# and a part of the window title may be given.

def matchTitle(wantedTitle, modInfo=None, titleExact=0, caseExact=0):
    """return progName if title matches, otherwise None
    """
    #pylint:disable=R0911    
    if modInfo is None:
        modInfo = natlink.getCurrentModule()
    if not (modInfo[0] and len(modInfo)) == 3:
        return None
    winTitle = modInfo[1]
    progName = getProgName(modInfo)
    
    if isinstance(wantedTitle, str):    
        if not caseExact:
            wantedTitle = wantedTitle.lower()
            winTitle = winTitle.lower()
        if titleExact:
            if winTitle == wantedTitle:
                return progName
            return False
        if winTitle.find(wantedTitle) >= 0:
            return progName
        return False
    if isinstance(wantedTitle, (list, tuple)):
        for t in wantedTitle:
            if matchTitle(t, modInfo, titleExact, caseExact):
                return progName
        return False
    print("unexpected place for matchTitle (unimacroutils)")
    return False

# Get basename of file:
def getBaseNameLower(name):
    """old fashioned way of getting the filename
    """
    return os.path.splitext(os.path.split(name)[1])[0].lower()


def getProgName(modInfo=None):
    """return the name of program (in the foreground)
    """
    if not modInfo:
        modInfo = natlink.getCurrentModule()
    return modInfo[0]

ProgInfo = collections.namedtuple('ProgInfo', 'progpath prog title toporchild classname hndle'.split(' '))

def getProgInfo(modInfo=None):
    """returns program info as namedtuple (progpath, prog, title, toporchild, classname, hndle)


    now length 6, including the classname, but also a named tuple!!

    prog always lowercase
    
    title now with capital letters.

    toporchild 'top' or 'child', or '' if no valid window
    """
    #pylint:disable=W0702
    if modInfo is None:
        try:
            modInfo = natlink.getCurrentModule()
        # print("modInfo through natlink: %s"% repr(modInfo))
        except natlink.NatError:
            progInfo = autohotkeyactions.getProgInfo()
            return progInfo
    
    _hndle = modInfo[2]
    if not _hndle:
        ## assume desktop, no foreground window, treat as top...
        return ProgInfo("", "", "", "top", "", 0)
    progpath = modInfo[0]
    prog = Path(modInfo[0].lower()).stem
    title = modInfo[1]
    if isTopWindow(modInfo[2]):
        toporchild = 'top'
    else:
        toporchild = 'child'
         
    HNDLE = modInfo[2]
         
    classname = win32gui.GetClassName(HNDLE)

    return ProgInfo(progpath, prog, title, toporchild, classname, HNDLE)
            
def getClassName(modInfo=None):
    """returns the class name of the foreground window
    take modInfo (tuple or int (the handle), or get it here)
    """
    if isinstance(modInfo, int):
        _hndle = modInfo
    elif isinstance(modInfo, tuple):
        _hndle = modInfo[2]
    else:
        _hndle = natlink.getCurrentModule()[2]
    if _hndle:
        return win32gui.GetClassName(_hndle)
    return None

def matchWindow(criteria, modInfo=None, progInfo=None):
    """looks for a matching window based on the dictionary of criteria

    criteria can either be a key with None as value, or a key
    with a part of window title as value (string), or a key with
    a list of parts of window titles as value.

    special these are:
    'all' (everything matches, value is ignored, so normally None)
    'none'  (nothing matches)
    'empty' (matches when no valid progInfo is found)

    progInfo is a tuple: (prog, title, topchild, hndle),
    prog being the lower case name of the programme
    title being the lower case converted title
    topchild being 'top' if top window, 'child' if child window,
                        if no valid module info

    progInfo may be omitted as well as modInfo.
        For best performance progInfo should be given,
        otherwise giving mtake unimacro couldodInfo is also faster than omitting it.

    """
    #pylint:disable=R0911, R0912
    #print 'matchwindow(qh): %s'% criteria
    if not isinstance(criteria, dict):
        print(f'type criteria in matchWindow function should be dict, not: {criteria}')
        return None
    
    if 'all' in criteria:
        return 1 
    if 'none' in criteria:
        return None

    _progpath, prog, title, topchild, _classname, _hndle = progInfo or getProgInfo(modInfo)
    if 'empty' in criteria and prog == '':
        return 1
    if 'top' in criteria and topchild != 'top':
        return None
    if 'child' in criteria and topchild != 'child':
        return None
    if prog in criteria:
        pot = criteria[prog]  # part of title
        if not pot:
            return 1   # no title given, so all titles match
        if isinstance(pot, str):
            return title.find(pot) >= 0   # one part of title given, check this
        if isinstance(pot, list):
            for t in pot:
                if title.find(t) >= 0:    # more possibilities for part of title
                    return 1
    elif 'top' in criteria or 'child' in criteria:
        return 1
    return None

## bringup functionality (moved from natlinkutilsbj to here):
pendingBringUps = []
pendingExecScripts = []

def doPendingBringUps():
    """called from _control"""
    #pylint:disable=W0603, W0702
    global pendingBringUps
    if not pendingBringUps:
        return
    for p in pendingBringUps:
        try:
            #print 'try to do pending bringup: %s'% p
            natlink.execScript(p)
        except:
            print('delayed bringup does not work: %s'% p)
    pendingBringUps = []
    
    
def doPendingExecScripts():
    """do pending execScripts, if there are any (from gotBegin or initialisation phase)
    """
    #pylint:disable=W0603, W0702    
    global pendingExecScripts
    if not pendingExecScripts:
        return
    for p in pendingExecScripts:
        try:
            #print 'try to do pending execScript: %s'% p
            natlink.execScript(p)
        except:
            print('does not work (in doPendingExecScripts): %s'% p)
    pendingExecScripts = []    

def ExecScript(script, callingFrom=None):
    """dall execScript, and put in pending list if fails (from gotBegin or initialisation phase)
    """
    #pylint:disable=W0603
    global pendingExecScripts
    if callingFrom and (getattr(callingFrom, 'status', '') == 'new' or
                        getattr(callingFrom, 'inGotBegin', 0)):
        print(f'pending execScript for {callingFrom.getName()}')
        pendingExecScripts.append(script)
        return None
                            
    try:
        natlink.execScript(script)
    except natlink.NatError as t:
        print('error in execScript: %s'% script)
        print('message: %s'% t)
        print('-------')
        return None
    return 1

def AppBringUp(App, Exec=None, Args=None, windowStyle=None, directory=None, callingFrom=None):
    """central BringUp function, can maintain old bringups
    BJ, extended QH.
    can be called from actions (UnimacroBringUp, or (do_)BRINGUP.

    """
    #pylint:disable=W0603, R0913, R0912
    global pendingBringUps
    app = str(App).lower()
    if Args:
        if isinstance(Args, list):
            args=' '.join(Args)
        else:
            args = Args
    else:
        args = None

##    if (GetOS()!='windows_nt'):
##        app=App.lower()
##    else:
##        app=App
    if not Exec:
        if args:
            cmdline = 'AppBringUp "%s", "%s"'% (app, args)
        else:
            raise UnimacroError("AppBringup should have at least Exec or Args (now: %s, %s)"%
                                (Exec, Args))
    else:
        ExecString = Exec
        # if os.path.isfile(Exec):
        #     if AppBringUpDoExec83:
        #         (fullName,ext)=os.path.splitext(Exec)
        #         if ext=='':
        #             Exec83=os.path.dirname(Exec)+'\\'+os.path.basename(Exec)
        #         else:
        #             Exec83=Exec
        #         ExecString = Exec83
        
        if args:
            # komodo, notepad (??)
            ExecString += " %s"% args
        cmdline = 'AppBringUp "%s","%s"'% (app, ExecString)

    if windowStyle:
        cmdline += ',%s'% windowStyle
    else:
        cmdline += ","
    if directory:
        cmdline += ', "%s"'% directory

    cmdline = cmdline.strip(", ")        
    #print "AppBringUp: %s"% cmdline

    if callingFrom and (getattr(callingFrom, 'status', '') == 'new' or
                        getattr(callingFrom, 'inGotBegin', 0)):
        #print 'pending AppBringUp for %s, %s'% (callingFrom.getName(), script)
        pendingBringUps.append(cmdline)
        return None
    try:
        ## this is a tricky thing, execScript only recognises str, not unicode!
        ## as default, ascii or cp1252 or latin-1 is taken, the windows defaults.
        _result = natlink.execScript(cmdline)
    except natlink.NatError as _t:
        #print 'wait for bringup until later: %s'% cmdline
        pendingBringUps.append(cmdline)
        return None
##    print 'ready withAppBringUp'
    return 1

# word formatting parameters:
wf_WordWasAddedByTheUser = 0x00000001
wf_InternalUseOnly1 = 0x00000002
wf_InternalUseOnly2 = 0x00000004
wf_WordCanNotBeDeleted = 0x00000008
wf_NormallyCapitalizeTheNextWord = 0x00000010
wf_AlwaysCapitalizeTheNextWord = 0x00000020
wf_UppercaseTheNextWord = 0x00000040
wf_LowercaseTheNextWord = 0x00000080
wf_NoSpaceFollowingThisWord = 0x00000100
wf_TwoSpacesFollowingThisWord = 0x00000200
wf_NoSpacesBetweenWordsWithThisFlagSet = 0x00000400
wf_TurnCapitalizationModeOn = 0x00000800
wf_TurnUppercaseModeOn = 0x00001000
wf_TurnLowercaseModeOn = 0x00002000
wf_TurnOffSpacingBetweenWords = 0x00004000
wf_RestoreNormalSpacing = 0x00008000
wf_InternalUseOnly3 = 0x00010000
wf_SuppressAfterAWordWhichEndsInAPeriod = 0x00020000
wf_DoNotApplyFormattingToThisWord = 0x00040000
wf_DoNotResetTheSpacingState = 0x00080000
wf_DoNotResetTheCapitalizationState = 0x00100000
wf_NoSpacePreceedingThisWord = 0x00200000
wf_RestoreNormalCapitalization = 0x00400000
wf_FollowThisWordWithOneNewLineCharacters = 0x00800000
wf_FollowThisWordWithTwoNewLineCharacters = 0x01000000
wf_DoNotCapitalizeThisWordInATitle = 0x02000000
wf_InternalUseOnly = 0x04000000
wf_AddAnExtraSpaceFollowingThisWord = 0x08000000
wf_InternalUseOnly4 = 0x10000000
wf_InternalUseOnly5 = 0x20000000
wf_WordWasAddedByTheVocabularyBuilder = 0x40000000

##wf_AddedInVersion8 = 0x20000000
##wf_DroppedInVersion9 = 0x60000000 ^ 0xffffffff
##print 'dropped: %x'% wf_DroppedInVersion9
#  List of word formatting properties as can be handled in
#    getWordInfo and setWordInfo
wordFormatting = {
    'WordWasAddedByTheUser': 0x00000001,
    'InternalUseOnly1': 0x00000002,
    'InternalUseOnly2': 0x00000004,
    'WordCanNotBeDeleted': 0x00000008,
    'NormallyCapitalizeTheNextWord': 0x00000010,
    'AlwaysCapitalizeTheNextWord': 0x00000020,
    'UppercaseTheNextWord': 0x00000040,
    'LowercaseTheNextWord': 0x00000080,
    'NoSpaceFollowingThisWord': 0x00000100,
    'TwoSpacesFollowingThisWord': 0x00000200,
    'NoSpacesBetweenWordsWithThisFlagSet': 0x00000400,
    'TurnCapitalizationModeOn': 0x00000800,
    'TurnUppercaseModeOn': 0x00001000,
    'TurnLowercaseModeOn': 0x00002000,
    'TurnOffSpacingBetweenWords': 0x00004000,
    'RestoreNormalSpacing': 0x00008000,
    'InternalUseOnly3': 0x00010000,
    'SuppressAfterAWordWhichEndsInAPeriod': 0x00020000,
    'DoNotApplyFormattingToThisWord': 0x00040000,
    'DoNotResetTheSpacingState': 0x00080000,
    'DoNotResetTheCapitalizationState': 0x00100000,
    'NoSpacePreceedingThisWord': 0x00200000,
    'RestoreNormalCapitalization': 0x00400000,
    'FollowThisWordWithOneNewLineCharacters': 0x00800000,
    'FollowThisWordWithTwoNewLineCharacters': 0x01000000,
    'DoNotCapitalizeThisWordInATitle': 0x02000000,
    'InternalUseOnly': 0x04000000,
    'AddAnExtraSpaceFollowingThisWord': 0x08000000,
    'InternalUseOnly4': 0x10000000,
    'InternalUseOnly5': 0x20000000,
    'WordWasAddedByTheVocabularyBuilder': 0x40000000,
                  
    }

def ListOfProperties(props):
    """return a list of word properties
    """
    l = []
    keyList = list(wordFormatting.keys())
    keyList.sort()
    for k in keyList:
        if props & wordFormatting[k]:
            l.append(k)
    return l
            

def makeWordProperties(listOfProps):
    """return the number that is the adding of the props

    in version 8 if word was added by user, add special to it
    in version 9 drop 0x60000000 from any props number
    """
    props = sys.maxsize
    for l in listOfProps:
        props += wordFormatting[l]
    props = sys.maxsize & props

# add to string 't' the count 'n'
#    eg   '{PgUp}', '3' --> '{PgUp 3}',
#         'abc', 3    --> 'abccc' 
#         'abc', '16'    --> 'ab{c 16}' 
def doCount(t,n):
    """add to string 't' the count 'n'

    eg   '{PgUp}', '3' --> '{PgUp 3}',
         'abc', 3    --> 'abccc' 
         'abc', '16'    --> 'ab{c 16}'
    """
    if len(t) == 0:
        return t
    if isinstance(n, str):
        n = int(n)
    if n <= 1:
        return t
    # do the work with braces:a}
    if hasBraces(t):
        if len(t) > 2 and not hasBraces(t[1:-1]):
            # balancing braces  { }:
            return "%s %s}"% (t[:-1], n)
        return t * n
    # only last char repeated:
    return t[:-1] + t[-1] * n

def hasBraces(t):
    """return True if t has braces
    """
    if not (t and isinstance(t, str)):
        return None
    first = '{' in t or '}' in t
    if first:
        if '{' in t[1:]:
            return None  # more than one character, treat as no braces
    return first

def doModifier(t,mod):
    """add Shift, Ctrl or Alt to the string 
       'a' wordt '{Ctrl+Shift+a}'
       '{Left} wordt '{Ctrl+Left}'
    """
    if len(t) == 0:
        return t
    if t.lower().find(mod.lower()) >= 0:
        return t # duplicate modifier?
    if t[0] == '{' and t[-1] == '}':
        return '{%s%s'% (mod, t[1:])
    return '{%s%s}'% (mod, t)

def doCaps(t):
    """make t uppercase if caps > 0
    
    strange old function
    """
    #pylint:disable=C0200
    for i in  range(len(t)):
        if t[i].islower():
            t = t[:i] + t[i].capitalize() + t[i+1:]
    return t
        
# mouseState to be altered when a downstate is asked for. Must be released before other actions
mouseState = 0 # button that is down (left = 1, right=2)
mouseModifier = 0 # could also be 'ctrl' or 'shift'(to do!)
# mouseStartPosition set by RM (rememberMouse()), returned to by CancelMouse()
mouseStartPosition = () # x, y

# mouse actions, called by MP and RMP from _commands.py)
# mouse = 'left' is most common (and default) input. If mouse = 0, '' of None
#   no clicking is done, if the mouse state = 0
#   if the mouseState = 1 (left) or 2 (right) dragging will continue.
#
# nClick = 1 is the normal input, if nClick = -1, the button is pushed down,
# if nClick = 0, the button is released

#
# for going to Joels routine, double work:
buttons = dict(noclick=0, left=1, right=2, middle=4)
joelsButtons = ['noclick', 'left', 'right', 'middle']
##print 'buttons: ', buttons

mouseDown = ['', natlinkutils.wm_lbuttondown, natlinkutils.wm_rbuttondown, natlinkutils.wm_mbuttondown]
mouseUp = ['', natlinkutils.wm_lbuttonup, natlinkutils.wm_rbuttonup, natlinkutils.wm_mbuttonup]

def doMouse(absorrel, screenorwindow, xpos, ypos, mouse='left', nClick=1, modifier = 0):
    """complicated, but complete mouse routine
    
    - absorrel: 0 = absolute positions (in pixels), 1 = relative positions (between -1.0 and 1.0)
    - screenorwindow: where to take the position from:
                0 = complete screen
                1 = active window
                2 = relative to the current position
                ## new 2017, relative to current monitor
                3 = relative to current monitor
                4 = relative to current monitor work area (excluding eg task bar and Dragon bar)
                5 = relative to the client area of or window (eg the body of an e-mail)
    - xpos, ypos: relative or absolute position horizontal and vertical
    - mouse: the button to be clicked: 'left', 'right', 1, 2, or 0 if no button is clicked
    - nClick: 1 or 2 OR -1 if the button has to be pushed down. 0 if the button must be released.


    -combined values for mouse for vocola MP and RMP calls:
    leftdouble, rightdouble, middledouble, leftrelease, rightrelease, middlerelease,
    leftup, rightup, middleup (same as release), and click (identical with left) and
    -also doubleclickleft or doubleleft etc.
    - noclick or move
    -abbreviations: down for leftdown, up for leftup release for leftrelease.
    
    Pushing down buttons down must be remembered, and reversed before 
    other actions are performed.  
    Therefore the mouse state is remembered in the variable mouseState.
    
    """
     #pylint:disable=W0603
    global mouseState
    hndle = natlink.getCurrentModule()[2]
    rect = None
    if hndle:
        rect = win32gui.GetWindowRect(hndle)
    xold,yold = natlink.getCursorPos()
    # screenorwindow corrensponds with relativeTo in DNS
    # only 0 (whole screen), 1 (relative to window), 2 (relative to cursorPos)
    # are considered here 5, inside client area
    if screenorwindow not in [0,1,2,3,4,5]:  # only within current window
        print("doMouse, only screenorwindow 0 ... 5 valid:", screenorwindow)
        return
    
    if screenorwindow == 0:
        _width, _height, xMin, yMin, xMax, yMax = monitorfunctions.getScreenRectData()
    elif screenorwindow == 1:
        if hndle:
            rect = win32gui.GetWindowRect(hndle)
        else:
            print("doMouse, no valid foreground window")
            return
        _width, _height, xMin, yMin, xMax, yMax = getRectData(rect)
    elif screenorwindow == 3:
        # get current monitor
        rect = monitorfunctions.get_current_monitor_rect( (xold, yold) )
        _width, _height, xMin, yMin, xMax, yMax = getRectData(rect)
    elif screenorwindow == 4:
        # get current monitor work area (skipping task bar and Dragon bar)
        rect = monitorfunctions.get_current_monitor_rect_work( (xold, yold) )
        _width, _height, xMin, yMin, xMax, yMax = getRectData(rect)
        # print 'w: %s, h: %s, xMin: %s, yMin: %s, xMax: %s, yMax: %s'% (
        #     width, height, xMin, yMin, xMax, yMax)
    elif screenorwindow == 5:
        if hndle:
            ## TODOQH???
            _clxold, _clyold = win32gui.ScreenToClient(hndle, (xold, yold) )
            rect = win32gui.GetClientRect(hndle)
        else:
            print("doMouse, no valid foreground window")
            return
        # active window
        _width, _height, xMin, yMin, xMax, yMax = getRectData(rect)

    if screenorwindow == 2:  # relative to current position
        xnew = xold + xpos
        ynew = yold + ypos
        xp, yp = monitorfunctions.get_closest_position( (xnew, ynew) )
    elif screenorwindow in (0, 1, 3, 4, 5): 
        if absorrel == 1: # relative:
            xp, yp = relToCoord(xpos, xMin, xMax), relToCoord(ypos, yMin, yMax)
        else:
            xp, yp = xpos, ypos   # get rid of: getMouseAbsolute(ypos, yMin, yMax)
        xp, yp = checkMousePosition(xp,xMin,xMax), checkMousePosition(yp,yMin,yMax)
        if screenorwindow == 5:
            clPos = (xp, yp)
            xp, yp = win32gui.ClientToScreen(hndle, clPos)
        # print 'before get_closest_position: (%s, %s)'% (xp, yp)
        xp,yp =  monitorfunctions.get_closest_position( (xp, yp) )
        # print 'after  get_closest_position: (%s, %s)'% (xp, yp)
    
    if debugMode == -1:
        print('Mouse to: %s, %s' % (xp, yp))
    nclick = 1
    onlyMove = 0
    if mouse and isinstance(mouse, str):
        # special variables for vocola combined calls:
        if mouse in ("noclick", "0"):
            nclick = 0
            mouse = ""
        elif mouse == 'move':
            onlyMove = 1
            nclick = 0
            mouse = ""
        else:
            # get the click and button from text:
            mouse, nclick = catchClick(mouse)
            if not nclick is None:
                nClick = nclick
            if not mouse:
                mouse = 'left'
    if mouse:
        if isinstance(mouse, str):
            if mouse not in buttons:
                print('doMouse warning: invalid value for mouse: %s (taking left button)'% mouse)
            btn = buttons.get(mouse, 1)
        else:
            btn = mouse
    else:
        #print 'no click, mouseState: %s'% mouseState
        btn = 0
        nclick = 0
    #nClick = nClick or nclick   # take things from "mouse" if nClick  not used
    #print 'btn: %s, nClick: %s, current mouseState: %s'% (btn, nClick, mouseState)
    if onlyMove:
        print('onlyMove to %s, %x'% (xp, yp))
        natlink.playEvents([(natlinkutils.wm_mousemove, xp, yp)])
    elif not mouseState:  # ongecompliceerd
        if nClick > 0:
            natlink.playEvents([(natlinkutils.wm_mousemove, xp, yp)])
            if btn:
                if debugMode == -1:
                    print('ButtonClick %s, %s' % (btn, nClick))
                else:
    ##                Wait()  # before clicking!
                    buttonClick(btn, nClick)
##                for i in range(nClick):
##                    print 'click'
##                    natlink.playEvents([(mouseDown[btn], xp, yp)])
##                    natlink.playEvents([(mouseUp[btn], xp, yp)])
##                    Wait(0.01)
                    
        elif nClick == 0:
            natlink.playEvents([(natlinkutils.wm_mousemove, xp, yp)])
        elif nClick == -1:
            if btn:
                if (xold,yold) != (xp, yp):
                    print('mousedown at old: %s, %s (%s)'% (xold, yold, repr(mouseDown[btn])))
                    natlink.playEvents([(mouseDown[btn], xold, yold)])

    ##            Wait()  # before clicking!
                print('mousedown at new: %s, %s (%s)'% (xp, yp, repr(mouseDown[btn])))
                natlink.playEvents([(mouseDown[btn], xp, yp)])
                mouseState = btn
    elif btn:  # muis was omlaag!
        #print 'btn: %s, wanted mouseState: %s'% (btn, mouseState)
        if mouseState != btn or nClick > 0: # change button
            if (xold,yold) != (xp, yp):
                natlink.playEvents([(mouseDown[mouseState], xold, yold)])
                natlink.playEvents([(mouseDown[mouseState], xp, yp)])

##            Wait()  # before clicking!
            natlink.playEvents([(mouseUp[mouseState], xp, yp)])
            mouseState = 0
            doMouse(0, 0, xp, yp, btn, nClick)
        # end or continue dragging state:
        if (xold,yold) != (xp, yp):
            natlink.playEvents([(mouseDown[btn], xold, yold)])
            natlink.playEvents([(mouseDown[btn], xp, yp)])
        if nClick == -1:
            pass
        else:
            natlink.playEvents([(mouseUp[btn], xp, yp)])
            mouseState = 0
    else:
        # no btn, but mouseState, simply move:
        natlink.playEvents([(natlinkutils.wm_mousemove, xp, yp)])  

def catchClick(mouse):
    """return reduced mouse command and nClick

    if mouse starts or ends with "doubleclick" or "double", nClick=2
    if mouse starts or ends with "click" , nClick=2
    if mouse starts or ends with "down", nClick=-1
    if mouse starts or endswith "up" or "release", nClick=0
    """    
    for clicker, result in [("doubleclick", 2), ("double", 2),
                            ("click", 1), ("down", -1),
                            ("up", 0), ("release", 0)]:
        if mouse.startswith(clicker) or mouse.endswith(clicker):
            mouse = mouse.replace(clicker, "")
            return mouse, result
    # fall through:
    return mouse, None 

##buttons = makedict(left=1, right=2, middle=4)
##joelsButtons = ['', 'left', 'right', 'middle']
def buttonClick(button='left', nclick=1, modifiers=None):
    """do a natspeak buttonclick, but release mouse first if necessary
    """
    # make button numeric:
    #if button in buttons:
    #    button = buttons[button]
    #if button not in [1, 2, 4]:
    #    raise UnimacroError('buttonClick invalid button: %s'% button)
    #if nclick not in [1,2]:
    #    raise UnimacroError('buttonClick invalid number of clicks: %s'% nclick)
    if mouseState:
        releaseMouse()
        
    natlinkutils.buttonClick(button, nclick, modifiers)
    #natlink.execScript("ButtonClick %s,%s"%(button, nclick))
    


def releaseMouse():
    """restores the default mouseState
    """
     #pylint:disable=W0603
    global mouseState
    if mouseState:
        (xp,yp) = natlink.getCursorPos()
        print('releasing mouse at %s, %s (%s)'% (xp, yp, repr(mouseUp[mouseState])))
        natlink.playEvents([(mouseUp[mouseState], xp, yp)])
        mouseState = 0
        Wait()

def mousePushDown(mouse='left'):
    """only pushes the mouse down

    """
    if mouseState:
        endMouse()
##    print 'pushing down with %s'% mouse
    doMouse(0, 2, 0, 0, mouse, -1)  # abs, rel to window, x, y, click
            
def endMouse():
    """replaced by releaseMouse"""
    releaseMouse()
        
def rememberMouse():
     #pylint:disable=W0603    
    global mouseStartPosition
    mouseStartPosition = natlink.getCursorPos()

def cancelMouse():        
     #pylint:disable=W0603    
    global mouseStartPosition
    if not mouseStartPosition:
        endMouse()
        print('cancelMouse, no mouseStartPosition')
        return
    if mouseState:
        doMouse(0, 0, mouseStartPosition[0], mouseStartPosition[1],
                mouseState, 0)
    else:
        doMouse(0, 0, mouseStartPosition[0], mouseStartPosition[1],
                0, 0)
    mouseStartPosition = ()
        
def checkMousePosition(pos, Min, Max):
    #print 'checking:', pos, Min, Max
    if pos >= Max: return Max - 1
    if pos < Min: return Min
    return int(pos)

def getRectData(rect):
    """get from a rect the width, height, xMin, xMax, yMin, yMax
    
    example: see unittestMouse.py (in unimacro_test directory)
    """
    xMin, yMin, xMax, yMax = tuple(rect)
    width, height = rect[2] - rect[0], rect[3] - rect[1]
    return width, height, xMin, yMin, xMax, yMax 
    
def coordToRel(x, xMin, xMax, side=0):
    """calculate relative coordinate, must be between 0 and 1
    side == 0: calculate from xMin (result positive) (default)
    side == 1: calculate from xMax (result negative)

    (can be used for x and y coordinates)
    
    example: see unittestMouse.py (in unimacro_test directory)
    """
    if not xMin <= x < xMax:
        return None # None means no result
    width = xMax - xMin
    if side == 0:
        return float(x-xMin)/width
    return float(x-xMax)/width


def relToCoord(relValue, xMin, xMax):
    """calculate coordinate, with relative value given
    
    if relValue >= 0 calculate from xMin
    if relValue < 0 calculate from xMax

    (can be used for x and y coordinates)
    
    example: see unittestMouse.py (in unimacro_test directory)
    """
    width = xMax - xMin
    if relValue >= 0:
        return int(xMin + float(relValue) * width + 0.5)
    return int(xMax + float(relValue) * width + 0.5)
   
#def getMouseRelative(rel, range, Min, Max):
#    """newer function, range is calculated inside function
#    """
#    return relToCoord(rel, Min, Max)

# def getMouseAbsolute(Pos, Min, Max):
#     """strange function was meant for positive Min and Max
#     
#     obsolete QH 2015
#     
#     with multiple screens it sucks
#     """
# 
#     if Pos < Min:
#         return Min
#     elif Pos > Max:
#         return Max
#     return Pos

cornerDict = {0:"top left", 1:"top right", 2:"bottom left", 3:"bottom right"}
whichDict = {0: "whole screen", 1:"active window", 3:"active monitor", 5:"client area"}
absorrelDict = {0: "absolute", 1:"relative"}

def getMousePositionActionString(absorrel, which, position):
    """return the proper action line (MP or RMP)
    parameter as in getMousePosition below
    
    if invalid, print lines in Messages window and return ""
    
    """
    mousePos = getMousePosition(absorrel, which, position)
    if mousePos is None:
        print(("current mouse position is invalid for a Unimacro Shorthand Command with parameters:\n"
              "absorrel: %s (%s), which: %s (%s), corner position: %s (%s)"% (absorrel, absorrelDict[absorrel],
                                             which, whichDict[which],
                                             position, cornerDict[position])))
        return ""
    x, y = mousePos
    if absorrel:
        return "RMP(%s, %.6s, %.6s)"% (which, x, y)
    return "MP(%s, %s, %s)"% (which, x, y)

def printMousePosition(absorrel, printAll = 0):
    """printing function for printing the mouse positions
    
    this function can be invoked by PMP and PRMP and PALLMP.
    
    these positions are recognised by the Unimacro Shorthand Commands MP and RMP
    """
    if printAll:
        print('-'*80)
        cornerRange = list(range(4))
    else:
        cornerRange = list(range(1))
    if absorrel:  # 1: relative:
        print('RELATIVE MOUSE POSITIONS:')
    else:
        print('ABSOLUTE MOUSE POSITIONS:')
    for which in whichDict:
        print('---related to %s:'% whichDict[which].upper())
        for cornerPos in cornerRange:
            print("%s: %s"% (cornerDict[cornerPos], getMousePositionActionString(absorrel, which, cornerPos)))
        if printAll:
            print('-'*20)

def getMousePosition(absorrel=0, which=0, position=0):
    """get the parameters for doMouse
    
    absorrel: 0 abs, 1 rel
    which: 1: active window, 5: client area, 0: whole screen
    position: 0: topleft,  1: topright, 2: bottomleft, 3: bottomright
    
    result:
    None if mouseposition is invalid for the choice in question (outside boundaries probably)
    
    otherwise a 2 tuple:
    x, y 
    """
    #pylint:disable=R0912, W0612 
    x, y = currentPos = natlink.getCursorPos()
    _hndle = natlink.getCurrentModule()[2]
    if which == 0:
        # complete screen
        width, height, xMin, yMin, xMax, yMax = monitorfunctions.getScreenRectData()
        if absorrel == 0:  # 0 absolute, just return position
            # screen absolute, only position:
            return x, y
    elif which == 1:
        # active window
        rect = win32gui.GetWindowRect(_hndle)
        width, height, xMin, yMin, xMax, yMax = getRectData(rect)
    elif which == 3:
        # active monitor
        rect = monitorfunctions.get_current_monitor_rect( currentPos )
        width, height, xMin, yMin, xMax, yMax = getRectData(rect)
        ##
    elif which == 5:
        # client area coordinates:
        x, y = win32gui.ScreenToClient(_hndle, currentPos)
        rect = win32gui.GetClientRect(_hndle)
        width, height, xMin, yMin, xMax, yMax = getRectData(rect)

    # now test for boundaries and abs or rel:
    if x < xMin or x > xMax or y < yMin or y > yMax:
        print('mouse position outside active window')
        return None
    if absorrel:  # 1: relative:
        #print 'RELATIVE MOUSE POSITIONS:'
        #print '---related to ACTIVE WINDOW:'
        if position == 0: # top left
            x, y = coordToRel(x, xMin, xMax, 0), coordToRel(y, yMin, yMax, 0)
        if position == 1: # top right
            x, y = coordToRel(x, xMin, xMax, 1), coordToRel(y, yMin, yMax, 0)
        if position == 2: # bottom left
            x, y = coordToRel(x, xMin, xMax, 0), coordToRel(y, yMin, yMax, 1)
        if position == 3: # bottom right
            x, y = coordToRel(x, xMin, xMax, 1), coordToRel(y, yMin, yMax, 1)
        return x, y
                    
    # else:        # absolute
    #print 'ABSOLUTE MOUSE POSITIONS:'
    #print '---related to ACTIVE WINDOW:'    
    if position == 0: # top left
        x, y = x-xMin, y-yMin
    if position == 1: # top right
        x, y = x-xMax, y-yMin
    if position == 2: # bottom left
        x, y = x-xMin, y-yMax
    if position == 3: # bottom right
        x, y = x-xMax, y-yMax
    return x, y
            

# gives true if the window is a "TOP" window (with a button on the
# windows task bar. False otherwise (it is a child window then)
def isTopWindow(hndle):
    """returns 1 if window is top,

    behaviour changed in python 2.3.4, parent = 0, no exception any more
    """
    #pylint:disable=W0702
    try:
        parent = win32gui.GetParent(hndle)
    except:
        return 1
    else:
        return parent == 0
        

# remember the current window, needed before waitForTitleChange can be done
# modInfo can be left away
Hndle = None
windowTitle = ""
def clearWindowHandle():
    """set to 0"""
    #pylint:disable=W0603    
    global Hndle
    Hndle = None

# This is called if the user clicks on the tray icon.  We simply cancel
# movement in all cases.
waitingCanceled = 0
iconState = 0
iconDirectory = os.path.join(status.getUnimacroDirectory(), 'icons')

# this sets the icontray for several waiting situations:::
def setTrayIcon(state=None, toolTip=None, comingFrom=None):
    """activate the trayIcon depending on the state
    
    switched off, reactivate later!!
    
    -If comingFrom is passed it can be either a instance (so the self of the calling grammar)
    or a function/method.
    -If comingFrom is an instance, a method "onTrayIcon" is taken (if possible) from the instance.
    In those cases this function (onTrayIcon) is called if the user clicks on the trayIcon.
    
    """
    #pylint:disable=W0101, R1710, W0603
    global iconState
    return None
    if state is None:
        natlink.setTrayIcon()
        return None # reset!
    #print 'natlinkqh setTrayIcon: %s'% state
    if state == 'waiting':
        toolTip = toolTip or 'unimacro is waiting'
        iconName = os.path.join(iconDirectory, 'waiting')
    else:
        iconName = state

    iconState = (iconState + 1)%2
##        print 'iconName: %s'% iconName
    if iconName[1] == ':':
        if not iconName.endswith('.ico'):
            # absolute path, attach extension:
            iconName = iconName + '.ico'
    elif iconName in ['right', 'left', 'up', 'down']:
        ## only if the 2 is not in the calling function yet...
        if iconState:
            iconName += '2'
    
    # if isinstance(comingFrom, instance):
    #     func = getattr(comingFrom, 'onTrayIcon', None)
    #     if func:
    #         toolTip += ' (grammar: %s)'% comingFrom.getName()
    # elif isinstance(comingFrom, func):
    #     func = comingFrom
    # elif comingFrom:
    #     func = None
    #     #print 'unimacroutils.setTrayIcon, comingFrom not of correct type (%s): %s'% (comingFrom, type(comingFrom))
    # else:
    #     func = None
    func = None    
    if func is None:
        toolTip += " (cannot be canceled)"
        try:
            natlink.setTrayIcon(iconName,toolTip)
        except natlink.NatError:
            natlink.setTrayIcon()
    else:
        try:
            natlink.setTrayIcon(iconName,toolTip,func)
        except natlink.NatError:
            print('cannot set tray icon "%s" (comingFrom: %s, func: %s), try to clear'% (iconName, comingFrom, func))
            natlink.setTrayIcon()
    return None
#         
# 
def clearTrayIcon():
    """clearTrayIcon, reactivate later
    """
    #pylint:disable=W0603
    global waitingCanceled, iconState
    waitingCanceled = 0
    iconState = 0
    # natlink.setTrayIcon()
# 

def rememberWindow(modInfo=None, progInfo=None, comingFrom=None):
    """remember the current window
    """
    #pylint:disable=W0603, W0613
    global Hndle, waitingCanceled, windowTitle
    # if not hndle true, then raised in error, because
    # rememberWindow is called before and not finished correct
    waitingCanceled = 0
    modInfo = modInfo or getCurrentModuleSafe()
    Hndle = modInfo[2]
    windowTitle = modInfo[1]
##    print 'set window to %s'% Hndle
    if not Hndle:
        print('warning, no window to remember: %s'% Hndle)
    # print 'rememberWindow, %s, %s'% (Hndle, windowTitle)
    return Hndle   

def waitForWindowTitle(titleName, nWait=10, waitingTime=0.1, comingFrom=None):
    """wait for window with title to come into the foreground

        waiting for a specific window title (word from...)
        all lowercase...
        timing out if nWait times is waited (eg 50 times), waitingTime suggested as 0.05 seconds
        (2.5 second total then!)

    """
##    print 'WWT, comingFrom: %s'% comingFrom
    titleName = titleName.lower()
    for _i in range(nWait):
##        print 'interrupted? %s'% comingFrom.interrupted
        if comingFrom and comingFrom.interrupted:
            # clearTrayIcon()
            print('waiting canceled (no TrayIcon handling')
            return None
        
        currentTitleName = natlink.getCurrentModule()[1].lower()
        # if empty (no window active) or matching:
        #print 'checking, currentTitle: %s, wantedTitle: %s'% (currentTitleName, titleName)
        if currentTitleName:
            if isinstance(titleName, str):
                if currentTitleName.find(titleName) >= 0:
                    # clearTrayIcon()
                    #print 'string test, found'
                    return 1
            else:
                for t in titleName:
                    if currentTitleName.find(t) >= 0:
                        # clearTrayIcon()
                        #print 'list test, found on  %s'% t
                        #found
                        return 1
        #print 'waiting, currentTitle: %s, wantedTitle: %s'% (currentTitleName, titleName)
        # setTrayIcon('waiting', comingFrom=comingFrom)
        Wait(waitingTime, comingFrom=comingFrom)
    # clearTrayIcon()
    print('Waiting for window title "%s" lasts too long, failed\nGot title: %s' % (titleName, currentTitleName))
    return None


def waitForNewWindow(nWait=10, waitingTime=0.1, comingFrom=None, debug=None, progInfo=None):
    """wait for a new window to appear in the foreground
    
    rememberWindow must run before. nWait and waitingTime as suggested above
    """
    #pylint:disable=R1710, W0613, W0612
    if Hndle is None:
        raise NatlinkCommandError("waitForNewWindow, no valid old hndle, do a rememberWindow() first")
    for i in range(nWait):
        if waitingCanceled:
            # clearTrayIcon()
            print('waiting canceled')
            return None

        modInfo = getCurrentModuleSafe()
        if not modInfo:
            print('waitForNewWindow failed, no modInfo')
            # clearTrayIcon()
            return None
        
        stepsToBeStable = max(3, i) # if it took longer to bring window in front, test more steps
        progN, titleN, hndleN = modInfo
        if hndleN != Hndle:
            # new window, wait for stable situation
            succes = 0
            for j in range(stepsToBeStable*3):
                # changed! reset wait a little and OK:
                Wait(waitingTime)
                newModInfo = getCurrentModuleSafe()
                if modInfo == newModInfo:
                    succes += 1
                    if succes >= stepsToBeStable:
                        break
                else:
                    succes = 0
                    modInfo = newModInfo
            else:
                print("waitForNewWindow: Found new window, but modInfo was not stable more than %s times"% stepsToBeStable)
                clearTrayIcon()
                return None
            if debug and j > stepsToBeStable + 1:
                extra = j - stepsToBeStable - 1
                print('title stable times after %s extra steps'% extra)
            clearTrayIcon()
            return 1
        setTrayIcon('waiting')
        Wait(waitingTime)
    clearTrayIcon()
    print("waiting for new window lasts too long, fail")
    return None

def waitForNewWindowTitle(nWait=10, waitingTime=0.1, comingFrom=None, progInfo=None):
    """wait for a new window with a matching title (???)
    """
    #pylint:disable=R1710, W0613
    if Hndle is None:
        raise NatlinkCommandError("waitForNewWindow, no valid old hndle, do a rememberWindow() first")
    for _i in range(nWait):
        if waitingCanceled:
            clearTrayIcon()
            print('waiting canceled')
            return None

        modInfo = natlink.getCurrentModule()
        if modInfo[1] and modInfo[1] != windowTitle:
            # changed! reset wait a little and OK:
            Wait(waitingTime)
            clearTrayIcon()
            return 1
        setTrayIcon('waiting')
        Wait(waitingTime)
    clearTrayIcon()
    print("waiting for new window title lasts too long, fail")
    return None
    
# return to window that was remembered by rememberWindow.
# nWait and waitingTime as suggested above.
def returnToWindow(nWait=5, waitingTime=0.05, winHandle=None, winTitle=None, **kw):
    """return to previous remembered window
    
    mostly do not specify winHandle and winTitle, as it is set as global
    variables hndle and windowTitle
    
    """
    winHandle = winHandle or Hndle
    winTitle = winTitle or windowTitle
    if not winHandle:
        print('returnToWindow, no window handle to return to, do nothing')
        return None
    if not win32gui.IsWindow(winHandle):
        print('returnToWindow, not a valid window: %s (%s)'% (winHandle, winTitle))
        return None
    # go:
    # print 'returning to window: %s (%s)'% (winHandle, winTitle)
    return SetForegroundWindow(winHandle, waitingTime=waitingTime, nWait=nWait)
    

# for returning a string, int or float:
restring = re.compile(r'[\'"].*[\'"]$')
reint = re.compile(r'[-]?[0-9]+$')
refloat = re.compile(r'[-]?[0-9]*[.][0-9]+$')

def arg2IntFloatOrString(arg):
    """utility function
    """
    #pylint:disable=C0321
    if reint.match(arg): return  int(arg)
    if refloat.match(arg): return  float(arg)
    return arg


def setDebugMode(n):
    """set the debug mode for commands that come from the "_commands" grammar
    """
    #pylint:disable=W0603    
    global debugMode
    debugMode = n

    
##def convertToNatSpeakArgString(input):
##    if not input:
##        return []
##    out = ' '
##    for i in input:
##        out = out + `i` + ', '
##    return out[:-2]


#  Wait some time, if no time is given,
#  wait 0.1 seconds = 100 milliseconds
def Wait(tt=None, comingFrom=None):
    """wait in milliseconds (version 5) or seconds (version 7)
    
    assume maximum time is 10 seconds, so if t > 10, milliseconds were given.
    
    between five and 10 a warning is given!
    
    no input: defaultWaitingTime is taken
    """
    t = tt or defaultWaitingTime
    if t > 10:
        print('warning, changed waiting time to seconds: %s'%t)
        t = t/1000.0
    elif t >= 5:
        print('warning, long waiting time: %s'% t)
        
    if debugMode == -1:
        print("Wait %s" % t)
    elif debugMode:
        t = t*debugMode
    if comingFrom is None:
        time.sleep(t)
    else:
        comingFrom.Wait(t)


def visibleWait():
    """wait less than defaultWaitingTime:
    """
    Wait(defaultWaitingTime*visibleWaitFactor)
    

def shortWait():
    """wait less than defaultWaitingTime:
    """
    Wait(defaultWaitingTime*shortWaitFactor)
    
def longWait():
    """wait 10 times the defaultWaitingTime:
    """
    Wait(defaultWaitingTime*longWaitFactor)

recharspace = re.compile('^[a-zA-Z- ]+$')
def addWordIfNecessary(w):
    """ see if word is already there, if not add.

    """    
    w = w.strip()
    if not w:
        return
    if not recharspace.match(w):
        print('invalid character in word to add: %s'% w)
        return
        
    isInVoc = (natlink.getWordInfo(w,1) is not None)
    isInActiveVoc = (natlink.getWordInfo(w,0) is not None)
    if isInActiveVoc:
        return
    try:
        if isInVoc:    # from backup vocabulary:
            print('make backup word active:', w)
            natlink.addWord(w,0)
            add2logfile(w, 'activated words.txt')
        else:
            print('adding word ', w)
            natlink.addWord(w)
            add2logfile(w, 'new words.txt')
            
    except natlink.InvalidWord:
        print('not added to vocabulary, invalid word: %s'% w)

def deleteWordIfNecessary(w):
    """delete word from vocabulary -- if needed
    """
    if not w:
        return
    isInActiveVoc = (natlink.getWordInfo(w,0) is not None)
    if isInActiveVoc:
        natlink.deleteWord(w)

if DEBUG:
    fOutName = 'c:\\DEBUG '+__name__+'.txt'
    debugFile = open(fOutName, 'w')
    print('DEBUG uitvoer naar: %s'% fOutName)

def debugPrint(t):
    """print to debug file (is this working???)
    """
    if not DEBUG:
        return
    if isinstance(t, str):
        debugFile.write(t)
    else:
        debugFile.write(repr(t))
    debugFile.write('\n')
    debugFile.flush()

def GetForegroundWindow():
    """return the handle of the current foreground window
    """
    return win32gui.GetForegroundWindow()
    

def SetForegroundWindow(h, waitingTime=0.1, nWait=3, debug=None):
    """gets the window in front
    
    Autohotkey is used if active!!
    
    When the switch is not made within 3 steps (of default waiting time),
    win+b (giving the system tray) is sent, and then the waiting cycle is done again a few times.
    
    tested in PyTest/unittestClipboard.py: testSwitchingWindows
    """
    #pylint:disable=R0912, C0321
    if not h:
        raise UnimacroError("no valid handle given for set foreground window: %s"% h)
    if autohotkeyactions.ahk_is_active():
        script = f'WinActivate, ahk_id {h}'
        ## TODOQH
        autohotkeyactions.do_ahk_script(script)
        # print(f'result SetForegroundWindow: {result}, wanted hndle: {h}')
        curHndle = win32gui.GetForegroundWindow()
        if curHndle == h:
            # print("autohotkey WinActivate succeeded: %s, wait 0.3 more seconds"% h)
            # time.sleep(0.3)
            return 1
        print(f'autohotkey did not get window with hndle {h} in the foreground')
        return None

    curHndle = win32gui.GetForegroundWindow()
    if curHndle == h:
        if debug: print('got it in one shot!! %s'% h)
        return 1
    
    if not win32gui.IsWindow(h):
        print('SetForegroundWindow: not a valid window hndle: %s'% h)
        return None
        
    for doKeystroke in [""]:  #  "{win+b}"]: ####, "{win+m}"]:
        result = _setForegroundWindow(h, doKeystroke, waitingTime=waitingTime, nWait=nWait, debug=debug)
        if result:
            if doKeystroke:
                if debug: print('SetForegroundWindow to %s success, after keystroke: %s'% (h, doKeystroke))
            else:
                if debug: print('SetForegroundWindow to %s success'% h)
            return result
        # if doKeystroke:
        #     if debug: 'return to calling window: %s'% curHndle
        #     win32gui.SetForegroundWindow(curHndle)
    return None          
    
def _setForegroundWindow(hndle, doKeystroke=None, waitingTime=0.1, nWait=3, debug=None):
    """try to switch to hndle
    """
    #pylint:disable=W0621, C0321, R0912
    if doKeystroke:
        if debug: print('try to get %s in foreground with keystroke: %s'% (hndle, doKeystroke))
        ## TODOQH make keystrokes here
        # natlinkutils.playString(doKeystroke)
    if win32gui.IsIconic(hndle):
        if debug: print('window %s is iconic, try to restore...')
        monitorfunctions.restore_window(hndle)
        Wait()
        for i in range(nWait):
            if win32gui.IsIconic(hndle):
                if debug and i:
                    print('window is still iconic, wait longer %s'% i)
                Wait(waitingTime)
            else:
                break
        else:
            # if debug: print('_setForegroundWindow, %s is still "Iconic"'% h)
            return None        
        
    win32gui.SetForegroundWindow(hndle)
  
    # except pywintypes.error as details:
    #     if details[0] in [0, 183]:
    #         pass
    #         # print 'could not bring to foreground: %s'% hndle
    #     else:
    #         raise
    for i in range(3):
        newH = getCurrentModuleSafe()[2]
        if newH == hndle:
            Wait()  #extra for safety
            return 1
        if debug and i:
            print("_setForegroundWindow, waiting %s (for %s, current: %s)"% (i, hndle, newH))
    if debug: print('_setForegroundWindow, with keystroke %s no result to set foregroundwindow to %s'% (doKeystroke, hndle))
    return None

titleHandles = {}
# switch to window with text in the title:
def switchToWindowWithTitle(wantedTitle, caseExact=0, titleExact=0):
    """keep cache of title handles for faster execution

    """
    #pylint:disable=W0603, C0321
    global titleHandles    
    extension = ''
    if titleExact: extension += "T"
    if caseExact: extension += "C"
    functionName = "lookForWindowText" + extension
    try:
        lookForFunction = globals()[functionName]
    except KeyError:
        print("invalid function in switchToWindowWithTitle: %s" % functionName)
        return None
    if len(titleHandles) > 100: # in case too many different switches
        print('clearing switchWindow title handles')
        titleHandles.clear()
    tryHandle = titleHandles.get(wantedTitle, 0)
    if tryHandle and not lookForFunction(tryHandle, wantedTitle):
        pass
    else:
        win32gui.EnumWindows(lookForFunction, wantedTitle)
        titleHandles[wantedTitle] = natlink.getCurrentModule()[2]
    Wait(0.1) # safety
    return 1
        
# look for window with part of text, case exact:
def lookForWindowTextC(hwnd, text):
    #pylint:disable=C0116
    if win32gui.GetWindowText(hwnd).find(text) >= 0:
        SetForegroundWindow(hwnd)
        return None
    return 1
# look for window Messages, default case, part of text, all to lower case:
def lookForWindowText(hwnd, text):
    #pylint:disable=C0116
    if win32gui.GetWindowText(hwnd).lower().find(text.lower()) >= 0:
        SetForegroundWindow(hwnd)
        return None
    return 1
# look for window with exact match:
def lookForWindowTextTC(hwnd, text):
    #pylint:disable=C0116
    if win32gui.GetWindowText(hwnd).strip() == text.strip():
        SetForegroundWindow(hwnd)
        return None
    return 1
# look for window Messages, whole title, but convert to lower case:
def lookForWindowTextT(hwnd, text):
    #pylint:disable=C0116
    if win32gui.GetWindowText(hwnd).lower().strip() == text.lower().strip():
        SetForegroundWindow(hwnd)
        return None
    return 1

def returnFromMessagesWindow():
    #pylint:disable=C0116    
    modInfo = natlink.getCurrentModule()
    if matchModule('natspeak', 'Messages from Python Macros', modInfo=modInfo):
        natlink.playString("{Alt+Tab}", natlinkutils.hook_f_systemkeys)

            
# Returns the date on a file or 0 if the file does not exist        
def getFileDate(fileName):
    #pylint:disable=C0116, C0321    
    try: return os.path.getmtime(fileName)
    except OSError: return 0        # file not found


def printListorString(arg):
    #pylint:disable=C0116    
    if isinstance(arg, str):
        print(arg)
    elif isinstance(arg, (list, tuple)):
        for l in arg:
            print(l)

      
#QH13062003  clipboard helper functions--------------------------------------

previousClipboardText = []

def saveClipboard():
    """Saves and clears the clipboard, and puts the text content in a global variable

    The global variable "previousClipboardText"
    is used to restore to the clipboard in the function "restoreClipboard"
    
    No input parameters, no result, the global variable is set

    """
    #pylint:disable=W0702    
    # global previousClipboardText
    t = getClipboard()
    print('clipboard contents to "previousClipboardText": %s'% t)
    previousClipboardText.append(t)
    for _i in range(10):
        try:
            win32clipboard.OpenClipboard()
            break
        except:
            time.sleep(0.1)
            continue
        else:
            print("could not open, save and empty the clipboard")
            return
    try:
        win32clipboard.EmptyClipboard()
    finally:
        win32clipboard.CloseClipboard()
    #print 'previousClipboardText: %s'% repr(previousClipboardText)

def clearClipboard():
    """clears the clipboard

    No input parameters, no result,

    """
    #pylint:disable=W0702
    # t0 = time.time()
    for _i in range(3):
        try:
            win32clipboard.OpenClipboard()
        except:
            print('error opening the clipboard')
            shortWait()
        else:
            break
    try:
        win32clipboard.EmptyClipboard()
    finally:
        win32clipboard.CloseClipboard()
    # print 'clearClipboard: %.4f'% (time.time() - t0,)

def restoreClipboard():
    """Restores the previously saved clipboard text into the clipboard

    No input, no result. The global variable is emptied.

    """
    #pylint:disable=W0702       
    # global previousClipboardText
    if previousClipboardText:
        t = previousClipboardText.pop()
    else:
        print('No "previousClipboardText" available, empty clipboard...')
        t = None
        return
    for _i in range(10):
        try:
            win32clipboard.OpenClipboard()
            break
        except:
            time.sleep(0.1)
            continue
        else:
            print("could not restore clipboard")
            return
    win32clipboard.EmptyClipboard()
    if t:
        #print 'restoring clipboard to: %s'% t
        win32clipboard.SetClipboardData(1, t)
    win32clipboard.CloseClipboard()

def getClipboard():
    """get clipboard through natlink, and strips off backslash r

    """
    #pylint:disable=W0702       
    # t0 = time.time()
    t = None
    for _i in range(3):
        try:
            t = natlink.getClipboard()
            if not t is None:
                # print ' at try', i
                break
        except:
            print('getClipboard, got no text')
            shortWait()
        else:
            break

    if t:
        # print 'got clipboard: "%s"'% repr(t)
        t = t.replace('\r', '')
        return t    
    print('getClipboard, got clipboard, but empty')
    return ''



def setClipboard(t, format=1):
    """set clipboard with text
    format = win32con.CF_UNICODETEXT (13): as unicode

    """
    #pylint:disable=W0622    
    win32clipboard.OpenClipboard()
    win32clipboard.EmptyClipboard()
    win32clipboard.SetClipboardData(format, t)
    win32clipboard.CloseClipboard()

def checkLists(one, two):
    """returns to lists, only in first, only in second

    if lists are equal 2 empty lists are returned
    """
    onlyone = []
    onlytwo = []
    for o in one:
        if not o in two:
            onlyone.append(o)
    for t in two:
        if not t in one:
            onlytwo.append(t)
        return onlyone, onlytwo

def cleanString(s):
    """converts a string with leading and trailing and
    intermittent whitespace into a string that is stripped
    and has only single spaces between words, newlines are removed.

>>> cleanString('foo  bar')
'foo bar'
>>> cleanString('\\n foo \\n\\n  bar ')
'foo bar'
    """
    return ' '.join([x.strip() for x in s.split()])



def cleanParagraphs(t):
    """make long paragraphs from selection, cleaning newlines and white space

    multiple empty lines are ignored, so spacing is 2 newlines at most.

    returns the new string    
    """
    T = t.split('\n\n')
    OUT = []
    hadEmptyLine = 0
    for p in T:
        if p.strip():
            OUT.append(cleanString(p))  
            hadEmptyLine = 0
        else:
            if not hadEmptyLine:
                OUT.append('')
            hadEmptyLine = 1
    return '\n\n'.join(OUT)
            

def stripSpokenForm(t):
    """strip the spoken from from a word from the Dragon Vocabulary
    """
    
    p = t.find('\\')
    if p == 0:
##        print 'backslash first: |%s|'% t
        if t[:4] in ['\\spa']:
            return ' '
        if t == '\\\\backslash':
            return '\\'
    elif p > 0:
        return t.split('\\', 1)[0]
    return t

# # get screen parameters:
# monitorfunctions.monitor_info()
# screenRect = monitorfunctions.VIRTUAL_SCREEN[:]
# screenWidth, screenHeight, screenXMin, screenYMin, screenXMax, screenYMax = getRectData(screenRect)
# logFolder = None
# try:
#     uud = getUnimacroUserDirectory()
#     if uud:
#         logFolder = os.path.join(uud, getLanguage() + "_log", getUser())
#         utilsqh.createFolderIfNotExistent(logFolder)
#         # print 'natlinkutilsqh, logfolder: %s'% logFolder
# except natlink.NatError:
#     pass
#     # print 'natlinkutilsqh, no logFolder active'


def add2logfile(word, filename):
    """add word to the filename in logFolder
    
    Disabled for now
    """
    #pylint:disable=E0602, W0702
    logFile = False
    if not logFile:
        return # silent
    try:
        f = open(os.path.join(logFolder, filename), 'a')
        f.write(word + '\n')
        print('written to %s: %s' % (os.path.join(logFolder), filename))
    except:
        pass
    
def _test():
    #pylint:disable=C0415
    import doctest
    return  doctest.testmod()

if __name__ == "__main__":
    _test()
    
    natlink.natConnect()
    _progInfo = getProgInfo()
    print("progInfo: %s"% repr(_progInfo))
    natlink.natDisconnect()
    


"""
This module tests the inivars module (the "unimacro inivars") with pytest

Quintijn Hoogenboom, 2021/2022 
"""
import time
from pathlib import Path
import pytest

# from dtactions import natlinkclipboard
# from dtactions.unimacro import unimacroinivars as inivars  # old style
from dtactions.unimacro import inivars

try:
    from dtactions.__init__ import getThisDir, checkDirectory
except ModuleNotFoundError:
    print('Run this module after "build_package" and "flit install --symlink"\n')
    raise

thisDir = getThisDir(__file__)
dtactionsDir = thisDir.parent

testDir = Path(thisDir)/'test_inivarsfiles'

def test_inifile():
    """testing a breaking inifile
    """
    testFile = '_brackets.ini'
    testPath = testDir/testFile
    ini = inivars.IniVars(testPath)
    pass
    
    
    
if __name__ == "__main__":
    pytest.main(['test_inivars.py'])



"""
This module tests the actions module (the "unimacro actions") with pytest

Quintijn Hoogenboom, 2021/2022 
"""
import time
from pathlib import Path
import pytest

# from dtactions import natlinkclipboard
# from dtactions.unimacro import unimacroactions as actions  # old style
from dtactions.unimacro import newactions as actions
from dtactions.unimacro import unimacroutils

try:
    from dtactions.__init__ import getThisDir, checkDirectory
except ModuleNotFoundError:
    print('Run this module after "build_package" and "flit install --symlink"\n')
    raise

thisDir = getThisDir(__file__)
dtactionsDir = thisDir.parent

dataDirDtactions = Path.home()/".dtactions"
dataDir = dataDirDtactions/'unimacro'
checkDirectory(dataDir) 

def test_simple():
    """only testing an empty action and T (true) and F (false), but with 1 as positive test result...Hello world
    """
    # this is when we go to class:
    # act = actions.Action()

    # now old module:Hello world
    act = actions
    result = act.doAction('T')
    assert result in (1, True)
    result = act.doAction('F')
    assert result is None
    
def xxx_test_keystroke():
    """test simple keystrokes and recover with clipboard
    """
    act = actions
    keystroke = act.doKeystroke
    keystroke("Hello world")
    keystroke('{shift+left 11}{ctrl+x}')
    time.sleep(0.01)
    result = unimacroutils.getClipboard()
    assert result == 'Hello world'
#
def test_save_return_clipboard():
    """test if the previous clipboard is collected afterwards
    """
    act = actions
    action = act.doAction
    keystroke = act.doKeystroke
    keystroke("{shift+up 3}{ctrl+c}")
    time.sleep(0.1)
    keystroke("{down 3}")
    start_text = unimacroutils.getClipboard()
    action("CLIPSAVE")

    keystroke("Hello world")
    keystroke('{shift+left 11}{ctrl+x}')
    time.sleep(0.1)
    result = unimacroutils.getClipboard()
    assert result == "Hello world"
    action("CLIPRESTORE")
    time.sleep(0.1)
    end_text = unimacroutils.getClipboard()
    assert start_text == end_text
    
if __name__ == "__main__":
    pytest.main(['test_actions.py'])



"""actions from application code (Visual Studio)

configuring instructions for Visual Studio at the bottom of this file

getting the current line number!
"""
import time
from dtactions.unimacro.actionclasses.actionbases import AllActions
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions.unimacro.unimacroactions import doAction as action
from dtactions import natlinkclipboard

class  CodeActions(AllActions):
    def __init__(self, progInfo):
        AllActions.__init__(self, progInfo)
        
    def getCurrentLineNumber(self, handle=None):
        debug - 0
        t1 = time.time()
        if self.toporchild == "child":
            return 0
        cb = natlinkclipboard.Clipboard(save_clear=True, debug=debug)  # clear "debug" to get rid of timing line
        # via the command palette:
        # action("{shift+ctrl+p}; copy current line to clipboard; {enter};")
          
        # better via a shortcut key, goto file, preferences, keyboard shortcuts
        # type copy current line number to clipboard
        # press ctrl+alt+c    (choose different keystroke and change in next line!!)
        shortcutkey = "{ctrl+alt+c}"
        keystroke(shortcutkey)
        
        # now collect the clipboard, at most waiting 10 intervals of 0.1 second.
        result = cb.get_text(waiting_interval=0.01, waiting_iterations=10)    # should be the current line number
        # print(f'result from clipboard: {result}')
        t2 = time.time()
        lapse = t2 - t1
        if debug:
            print(f'time in getCurrentLineNumber: {lapse:.3f}')
        try:
            return int(result)
        except (ValueError, TypeError):
            return 0
        
    
if __name__ == '__main__':
    pass        
        
# Note, the program name of Visual Studio Code is "code.exe", so this module is called code.py
#
# Please install the VS Code extension called 'copy-current-line-number'
# that copies the current editor line number to the clipboard.
#
# Next go to the File, Preferences, Keyboard Shortcuts.
# lookup your extension (copy current line number), and attach ctrl+alt+c to this extension.

# After you configure the grammar _lines.py, like:
# [general]
# line numbers modulo hundred = T
# and toggle the microphone, possibly restart Dragon,
# the feature should work.
#
# each line number below 100, will jump to the nearest line with respect to the current line.
        
        


"""actions from application uedit32, UltraEdit

(from IDM Computer Solutions, www.ultraedit.com)
"""
import win32gui
import ctypes
import win32api
import unimacro.messagefunctions
import as
import mf
from .actionbases import MessageActions
import pprint

class Uedit32Actions(MessageActions):
    def __init__(self, progInfo):
        MessageActions.__init__(self, progInfo)
        
    def getInnerHandle(self, handle):
        """procedure to get inner handle for EditControl,
        
        via a control that holds the filename, which is extracted from the window title
        """
        title = win32gui.GetWindowText(handle)
        #print 'title: %s'% title
        tabName = title.split(']')[0]
        tabName = tabName.split('[')[-1]
        self.prevTabName = tabName
        #print 'tabName: %s'% tabName
        controls = mf.findControls(handle, wantedText = tabName)
        if not controls:
            #print 'uedit32Actions, no control with title: "%s" found'% tabName
            return
        if len(controls) > 1:
            print 'uedit32Actions, strange, found more controls, matching the title: "%s", take first'% tabName
        c = controls[0]
        innerControl = mf.findControls(c, wantedClass="EditControl")[0]
        return innerControl
    
if __name__ == '__main__':

    # test, fill in handle of ultraedit (uedit32) (with "give window info" from grammar _general)
    focus =199852
    print 'focus; %s'% focus
    progInfo = ('uedit32', 'babbababa', 'top', focus)
    ue = Uedit32Actions( progInfo )



"""unimacro natlink macro wrapper/extensions
(c) copyright 2003 Quintijn Hoogenboom (quintijn@users.sourceforge.net)

 autohotkeyactions.py 
 make it independent from Natlink (Dragon), but it is windows specific

 written by: Quintijn Hoogenboom (QH softwaretraining & advies)
 December 2013/.../March 2021, adding AutoHotkey support

This module contains actions via AutoHotkey

(see autohotkey.com)

1. the ahkexe and ahkscriptfolder are checked for at the bottom of the import procedure.
   Previous scripts are copied from the directory dtactions/samples/autohotkey

2. ahk_is_active returns whether AutoHotkey is running on your system. 

3. call_ahk_script_path calls the script you have defined.

4. If you have the script in a string, call call_ahk_script.
    The script is copied to tempscript.ahk in the ahkscriptfolder and executed with call_ahk_script_path
    
5. When scripts want information back into python, write this information into a file eg "INFOfromAHK.txt".
   Scripts that use this feature, should have a function in this module, in order to read the data from this file.
   See getProgInfo and getCurrentModule below.

timing results:
  1    0.000    0.000    0.401    0.401 autohotkeyactions.py:456(killWindow)
  4    0.000    0.000    0.261    0.065 autohotkeyactions.py:91(getProgInfo)
  1    0.000    0.000    0.234    0.234 autohotkeyactions.py:209(ahkBringup)
  1    0.000    0.000    0.052    0.052 autohotkeyactions.py:323(GetForegroundWindow)
  
"""
import subprocess
import shutil
import filecmp
from pathlib import Path
import collections
import os.path  # only for getting the ahk exe path...
import time
from textwrap import dedent
from dtactions.sendkeys import sendkeys
## get thisDir and other basics...
try:
    from dtactions.__init__ import getThisDir, checkDirectory
except ModuleNotFoundError:
    print('Run this module after "build_package" and "flit install --symlink"\n')
    raise


dtactions = thisDir = getThisDir(__file__)
sampleAhkDirectory = dtactions/'samples'/'autohotkey'
checkDirectory(sampleAhkDirectory)

ahkexe = None
ahkscriptfolder = None

def ahk_is_active():
    """return True if autohotkey is active
    """
    return bool(ahkexe) and bool(ahkscriptfolder)

def do_ahk_script(script, filename=None):
    """try autohotkey integration
    """
    filename = filename or 'tempscript.ahk'
    if not ahk_is_active():
        print('ahk is not active, cannot run script')
        return
    #print 'AHK with script: %s'% script
    scriptPath = ahkscriptfolder/filename
    if isinstance(script, (list, tuple)):
        script = '\n'.join(script)         
    if not script.endswith('\n'):
        script += '\n'
    
    with open(scriptPath, 'w') as fp:
        fp.write(script)
    call_ahk_script_path(scriptPath)

def call_ahk_script_path(scriptPath):
    """call the specified ahk script  
    
    you can call .ahk scripts or .exe compiled files,
    but .exe files seem not to give better performance...
    
    use the global variable ahkexe as executable
    
    """
    scriptPath = str(scriptPath)
    if scriptPath.endswith('.ahk'):
        result = subprocess.call([ahkexe, scriptPath, ""])
    elif scriptPath.endswith('exe'):
        result = subprocess.call([scriptPath, "", ""])
    else:
        raise ValueError(f'autohotkeyactions, call_ahk_script_path: path should end with ".ahk", or ".exe"\n    path: {scriptPath}')
    if result:
        print('non-zero result of call_ahk_script_path "%s": %s'% (scriptPath, result))

ProgInfo = collections.namedtuple('ProgInfo', 'progpath prog title toporchild classname hndle'.split(' '))

def getProgInfo():
    """get the prog info, like natlink.getCurrentModule enhanced with toporchild and classname
    
    returns program info as namedtuple (progpath, prog, title, toporchild, classname, hndle)

    So the length of progInfo is 6!

    toporchild 'top' or 'child', or '' if no valid window
          
    """
    _ErrorFile = clearErrorMessagesFile()
    ProgInfoFile = clearProgInfoFile()
    
    script = getProgInfoScript(info_file=ProgInfoFile)
    do_ahk_script(script)
    
    return getProgInfoResult(info_file=ProgInfoFile)

GetProgInfo = getProgInfo

def getProgInfoScript(info_file):
    """return the ahk script for getting the progInfo
    """
    script = """
; put prog info of current window in file.
;(progpath, prog, title, toporchild, classname, hndle)

WinGet pPath, ProcessPath, A
WinGetTitle, Title, A
WinGetClass, Class, A
WinGet wHndle, ID, A
wHndle := wHndle + 0
toporchild := "top"
if DllCall("GetParent", UInt, WinExist("A")) {
    toporchild := "child"    
}

FileDelete, ##proginfofile##
FileAppend, %pPath%`n, ##proginfofile##
FileAppend, %Title%`n, ##proginfofile##
FileAppend, %toporchild%`n, ##proginfofile##
FileAppend, %Class%`n, ##proginfofile##
FileAppend, %wHndle%, ##proginfofile##
"""
    script = script.replace('##proginfofile##', str(info_file))
    return script

def getProgInfoResult(info_file):
    """extract the contents of the info_file, and return the progInfo
    """
    with open(info_file, 'r') as fp:
        progInfo = fp.read().split('\n')
        
    # note ahk returns 5 lines, but ProgInfo has 6 items.   
    if len(progInfo) == 5:
        # make hndle decimal number:
        progpath, title, toporchild, classname, hndle = progInfo
        if hndle:
            hndle = int(hndle)
        else:
            hndle = 0
        prog = Path(progpath).stem
        return ProgInfo(progpath, prog, title, toporchild, classname, hndle)
    raise ValueError(f'ahk script getProgInfo did not return correct result:\n    length {len(progInfo)}\n    text: {repr(progInfo)}')


#def call_ahk_script_text(scriptText):
#    """call the specified ahk script as a text string
#    
#    use the com module, does not work (yet)
#    
#    """
#    script = scriptText
#    if not script.strip().endswith("ExitApp"):
#        script += "\nExitApp"
#    print 'do script via dll: %s'% script
#    ahk = win32com.client.Dispatch("AutoHotkey.Script")
#        
#    ahk.ahktextdll(script)
 
def copySampleAhkScripts(fromFolder, toFolder):
    """copy (in new Autohotkey directory) the sample script files
    """
    if not fromFolder.is_dir():
        print(f'No sample_ahk dir found (should be in dtactions/sample): {fromFolder}')
        return
    for f in fromFolder.glob("*.ahk"):
        inputFile = f
        inputName = f.name
        # stem = f.stem
        outputFile = toFolder/inputName
        if not outputFile.is_file():
            print(f'---copy AutoHotkey script "{inputName}" from\nSamples directory to "{toFolder}"----')
            shutil.copyfile(inputFile, outputFile)
        elif getFileDate(inputFile) > getFileDate(outputFile):
            if not filecmp.cmp(inputFile, outputFile): # inputFile (samples), newer and changed
            #     for i in range(1000):
            #         outputStem = f'{stem}_{i:03d}'
            #         newOutputFile = outputFile.with_name(newOutputName)
            #         if not newOutputFile.is_file():
            #             break
            #     else: 
            #         raise OSError('no unused newOutputFile available last: {newOutputFile}')
            #     print(f'AutoHotkey script "{inputName}" has been changed in "sample_ahk"\n   copy to "{toFolder}"\n   keep backup in {newOutputFile}')
            #     shutil.copyfile(outputFile, newOutputFile)
                shutil.copyfile(inputFile, outputFile)
        
def ahkBringup(app, filepath=None, title=None, extra=None, waitForStart=1):
    """start a program, folder, file, with AutoHotkey, or bring to foreground
    
    This functions is related to UnimacroBringup, which works with AppBringup from Dragon,
    but sometimes fails.
    
    Besides, this function can also work without Dragon being on, not relying on natlink.
    So better fit for debugging purposes.
    
    app:               can be without .exe (so notepad or notepad.exe), or the complete path
                       of the app. Advised when you want to activate files in a specific app,
                       that is not the default app for the filetype.
    filepath:          optional, the file to open.
    title:             activate title if the app already exists
    extra:             do extra actions after finding the correct app in the foreground,
                       such as opening a new document
    waitForStart:      (default: 3), wait longer for a new process to start (in seconds)
    waitForActivate:   obsolete, not needed in practice
    
    currently appBringup with app and title seems to be pretty stable, tested with thunderbird.
    other variants need testing
    
    returns: progInfo if ahkBringup succeeded
    returns: error message if ahkBringup failed

    error messages (str):
        1. when in python code, return direct
        2. when raising an error in the ahk script, put text in errormessagefromahk.txt
        
    Uses two text files for getting the result:
        1. errormessagefromahk.txt: retrieves error messages as reported from the ahk script
        2. proginfofromahk.txt: gives the progInfo of the window that has been brought up
    """
    #pylint:disable=R0913, R0911, R0912
    if not ahk_is_active():
        return 'cannot run ahkBringup, autohotkey is not active'
    ErrorFile = clearErrorMessagesFile()
    ProgInfoFile = clearProgInfoFile()
    
    title  = title or ''
    filepath = filepath or ''
    extra = extra or ''
    # if app.endswith('.exe'):
    #     basename = app[:4]
    # else:
    #     basename = app
    # app = app + '.exe'

    
    ## treat mode = open or edit, finding a app in actions.ini:
    if ((app and app.lower() == "winword") and
        (filepath and (str(filepath).endswith(".docx") or str(filepath).endswith('.doc')))):
        if Path(filepath).is_file():
            script = _get_run_winword_script(filepath)
        else:
            return f'ahkBringup winword with filepath: "{filepath}": does not exist'

    else:
        # get function and vars
        Func, Vars = _get_run_function_vars(app, filepath, title)
        if Func is None:
            return f'ahkBringup, invalid combination of input variables\n    app: "{app}"\n    filepath: "{filepath}"\n    title: "{title}"'
        script = Func(*Vars)

    if isinstance(script, str):
        script = [script]
    # script.append("WinWait, ahk_pid %NewPID%") ## this is done in the Func above
    # now grab all
    # why this one???
    # script.append("WinGet, pPath, ProcessPath, ahk_pid %NewPID%")
    if extra:
        script.append(extra)
    # always:    
    script.append(getProgInfoScript(info_file=ProgInfoFile))
    script = '\n'.join(script)
    ## do the replacements:
    script = script.replace('##waitForStart##', str(waitForStart))
    script = script.replace('##proginfofile##', str(ProgInfoFile))
    script = script.replace('##ErrorFile##', str(ErrorFile))

    ## do the script!!
    do_ahk_script(script)
    message = open(ErrorFile, 'r').read().strip()
    if message:
        return message  

    # see if there is an error message:
    errors = readErrorMessagesFile()
    if errors:
        return errors

    progInfo = getProgInfo()
    ## collect the progInfo:
    if len(progInfo) != 6:
        return 'progInfo returns tuple of invalid length (should be 6): {len(progInfo)}\n===={progInfo})'
    if not progInfo.hndle:
        return 'autohotkeyactions.appBringup, no valid progInfo returned, hndle is empty'
    if progInfo and isinstance(progInfo, str):
        ## error message:
        return progInfo
    ## OK!:
    return progInfo

autohotkeyBringup = ahkBringup

## here the internal functions for appBringup:
def _get_run_winword_script(filepath):
    """construct script than opens an existing word document
    
    filepath is the Word document to open.

    The retrieval of the progInfo is done in the appBringup function    
    """
    script = '''\
        Word := ComObjCreate("Word.Application")
        Word.Visible := True
        Word.Documents.Open("##filepath##")
        Word.Visible := 1
        Word.Activate
        if ErrorLevel {
            FileAppend, ahkBringup: could not start winword with document ##filepath##te, ##ErrorFile##
            return
        }
    '''
    script = script.replace('##filepath##', filepath)
    script = dedent(script)
    return script

def _get_run_function_vars(app, filepath, title):
    """return _app_filepath or _app_title or _app, dependent on the variables being True
 
    This is the script that is tested with thunderbird in testAhkBringupThunderbirdAppTitle

    input: values of app, filepath and title
    output: callable function or None, tuple of input variables
    
>>> _get_run_function_vars('notepad', '', '')               #doctest: +ELLIPSIS
(<function _get_run_app_script at ...>, ('notepad',))
>>> _get_run_function_vars('notepad', '', 'window title')   #doctest: +ELLIPSIS
(<function _get_run_app_title_script at ...>, ('notepad', 'window title'))

# wrong call:
>>> _get_run_function_vars('', '', '')   #doctest: +ELLIPSIS
(None, ())
    
    """
    Parts = []
    Vars = []
    if app:
        Vars.append(app)
        Parts.append('_app')
    if filepath:
        Parts.append('_filepath')
        Vars.append(filepath)
    if title:
        Parts.append('_title')
        Vars.append(title)
    func_name = f'_get_run{"".join(Parts)}_script'
    if func_name in globals():
        func = globals()[func_name]
    else:
        func = None
    return func, tuple(Vars)

def _get_run_app_title_script(app, title):
    """constructs start of script switching to window title if present otherwise open app
    
    """
    script = '''\
        SetTitleMatchMode, 2
        ; _get_run_app_title_script, app: "##app##", title: "##title##" 
        wantedHndle:= WinExist(##title##)
        WinGet, activeHndle, ID, A
        
        if (wantedHndle) {
            if (activeHndle != wantedHndle) {
                if WinExist(##title##)
                    WinActivate
            }
        }
        else {
            Run, ##app##,,,NewPID
            WinWait ahk_pid %NewPID%,,##waitForStart##
            if ErrorLevel {
                FileAppend, Could not activate app ##app##, ##ErrorFile##
                return
                    }
        }
        ; end of _get_run_app_title_script part of script
        '''
    script = script.replace('##app##', f'"{app}"')
    script = script.replace('##title##', f'"{title}"')
    script = dedent(script)
    return script

def _get_run_app_filepath_script(app, filepath):
    """constructs start of script switching running app with a filepath
    """
    script = [f'Run, "{app} {filepath}",,, NewPID']
    errorlines = '''
        if ErrorLevel {
            FileAppend, Could not activate app: "##app##" with filepath: "##filepath##", ##ErrorFile##
            return
            }
        '''
    errorlines = errorlines.replace('##app##', app)
    errorlines = errorlines.replace('##filepath##', filepath)
    errorlines = dedent(errorlines)
        
    script.append(errorlines)
    return script

def _get_run_filepath_script(filepath):
    """constructs start of script switching running a filepath
    """
    script = [f'Run, "{filepath}",,, NewPID']
    errorlines = '''
        if ErrorLevel {
            FileAppend, Could not start filepath: "##filepath##", ##ErrorFile##
            return
            }
        '''
    errorlines = errorlines.replace('##filepath##', str(filepath))
    errorlines = dedent(errorlines)
    script.append(errorlines)
    return script

def _get_run_app_script(app):
    """constructs start of script switching running an app
    """
    script = [f'Run, "{app}",,, NewPID']
    errorlines = '''
        if ErrorLevel {
            FileAppend, Could not start app: "##app##", ##ErrorFile##
            return
            }
        '''
    errorlines = errorlines.replace('##app##', app)
    errorlines = dedent(errorlines)
    script.append(errorlines)
    return script

def GetForegroundWindow():
    """return the hndle of the ForegroundWindow
    
    OK: return value, int, the hndle of the Foreground window
    Error: return value is str: error message explaining the failure
    """
    HndleFile = ahkscriptfolder/"foregroundhndlefromahk.txt"
    
    script = '''\
        ; put window hndle of current window in file.
        WinGet pPath, ProcessPath, A
        WinGet wHndle, ID, A
        wHndle := wHndle + 0
        FileDelete, ##hndlefile##
        FileAppend, %wHndle%, ##hndlefile##
    '''
    script = script.replace('##hndlefile##', str(HndleFile))
    do_ahk_script(script, filename="getforegroundwindow.ahk")

    with open(HndleFile, 'r') as fp:
        gotHndle = fp.read().strip()
    try:
        if gotHndle:
            hndleInt = int(gotHndle)
            return hndleInt
        mess = 'autohotkeyactions, GetForegroundWindow did not return a window hndle'
        return mess
    except ValueError:
        if gotHndle:
            mess = f'autohotkeyactions, GetForegroundWindow: did not get correct hndle window: {gotHndle}'
            # print(mess)
            return mess
        mess = 'autohotkeyactions, GetForegroundWindow: did not get correct hndle window: (empty)'
        # print(mess)
        return mess

def SetForegroundWindow(hndle):
    """bring window with hndle into the foreground
     
    return True: success
    other return: message why it went wrong
    
    """
    ErrorFile = clearErrorMessagesFile()
    ProgInfoFile = clearProgInfoFile()

    script = '''\
        WinActivate, ahk_id ##hndle##
        WinWaitActive, ahk_pid ##hndle##,,1
        if ErrorLevel {
            FileAppend, Could not get window with hndle ##hndle## in the foreground, ##ErrorFile##
            return
        }

        WinGet wHndle, ID, A
        wHndle := wHndle + 0
        FileDelete, ##proginfofile##
        FileAppend, %wHndle%, ##proginfofile##
        '''
    script = script.replace('##hndle##', str(hndle))
    script = script.replace('##proginfofile##', str(ProgInfoFile))
    script = script.replace('##ErrorFile##', str(ErrorFile))
    do_ahk_script(dedent(script))
    
    if Path(ErrorFile).is_file():
        mess = f'Error with SetForegroundWindow to {hndle}'
        return mess

    if not Path(ProgInfoFile).is_file():
        mess = f'Error with SetForegroundWindow to {hndle}, InfoFile cannot be found'
        return mess
        
    winHndle = open(ProgInfoFile, 'r').read().strip()
    if winHndle:
        try:
            winHndle = int(winHndle)
        except ValueError:
            mess = f'ahk script getIntoForeground to "{hndle}" did not return correct winHndle: {winHndle}'
            return mess
            
        if winHndle == hndle:
            return True  # ok, then return True
        mess = f'could not switch to wanted hndle {hndle} (got {winHndle})'
        return mess
    raise ValueError('ahk script getIntoForeground did not return anything')

def getFileDate(modName):
    """return the last modified date/time of file, 0 if file does not exist
    """
    try:
        modTime = modName.stat().st_mtime
        return modTime
    except OSError:
        return 0        # file not found

def GetAhkExe():
    """try to get executable of autohotkey.exe, if not there, empty string is put in ahkexe
    
    return None, raises OSError if ahkExe is not found
    """
    # pylint: disable=W0603  # (using the global statement)
    global ahkexe
    # no succes, go on with program files:
    pf = os.path.expandvars("%PROGRAMFILES%")
    if pf.find('(x86)')>0:
        # 64 bit:
        pf = os.path.expandvars("%PROGRAMW6432%") 
    if pf and Path(pf).is_dir():
        pf = Path(pf)
    elif pf:
        raise OSError('cannot find (old style) program files directory: (empty)')
    else:
        raise OSError(f'cannot find (old style) program files directory: {pf}')
    
    ahk = pf/"autohotkey"/"autohotkey.exe"
    if ahk.is_file():
        ahkexe = ahk
        #print 'AutoHotkey found, %s'% ahkexe
    else:
        ahkexe = ""
        print(f'AutoHotkey not found on this computer {ahkexe}')
    
def GetAhkScriptFolder():
    """try to get AutoHotkey folder as subdirectory of HOME
    
    On Windows mostly C:\\Users\\Username\\.autohotkey
    
    create if non-existent.
    """
    # pylint: disable=W0603  # (using the global statement)
    global ahkscriptfolder

    if ahkexe is None:
        GetAhkExe()
    if not ahkexe:
        ahkscriptfolder = ""
        return ahkscriptfolder

    if not ahkscriptfolder is None:
        ## for repeated use:
        return ahkscriptfolder

    scriptfolder = Path.home()/".autohotkey"
    checkDirectory(scriptfolder, create=True)
    ahkscriptfolder = scriptfolder
    copySampleAhkScripts(sampleAhkDirectory, ahkscriptfolder)
    return scriptfolder

def clearProgInfoFile():
    """remove the previous ProgInfo file if present
    
    return the path of the ProgInfoFile
    """
    ProgInfoFile = ahkscriptfolder/"proginfofromahk.txt"
    if ProgInfoFile.is_file():
        ProgInfoFile.unlink()
    return ProgInfoFile

def clearErrorMessagesFile():
    """make an empty ErrorMessagesFile
    
    return the path of the ErrorMessagesFile
    """
    ErrorFile = ahkscriptfolder/"errormessagefromahk.txt"
    with open(ErrorFile, 'w') as f:
        f.write('\n')
    return ErrorFile

def readErrorMessagesFile():
    """get the error messages if any
    """
    ErrorFile = ahkscriptfolder/"errormessagefromahk.txt"
    with open(ErrorFile, 'r') as f:
        mess = f.read()
    if mess.strip():
        return f'autohotkeyactions.ahkBringup failed:\n===={mess}'
    return ''



def killWindow(hndle=None, key_close=None, key_close_dialog=None, silent=True):
    """kill the app with hndle,
    
    like the unimacro shorthand command KW (killwindow)
        
    input:
    hndle: int, windows handle of the app to be closed, normally leave away,
                the foreground hndle is taken
    key_close: keystrokes to be performed to close a top window, default: `{alt+f4}`
    key_close_dialog: keystrokes to close the dialog if a dialog (child window) is in
                      the foreground. Default  `{alt+n}`
    silent: if messages are printed (True by default)
    
    returns: True if success
             message with progInfo (if possible) if fail
    
    tested with unittestAutohotkeyactions.py...
    """
    # pylint: disable=R0911, R0912
    _ErrorFile = clearErrorMessagesFile()
    _ProgInfoFile = clearProgInfoFile()
   
    if hndle:
        result = SetForegroundWindow(hndle)
        if result is not True:
            mess = f'window {hndle} not any more available'
            if not silent:
                print(mess)
            return mess
    progInfo = getProgInfo()
    if hndle:
        if hndle != progInfo.hndle:
            mess = f'hndle {hndle} does not match hndle of foreground window\n{progInfo}'
            return mess
    else:
        hndle = progInfo.hndle
    
    key_close = key_close or "{alt+f4}"
    key_close_dialog = key_close_dialog or "{alt+n}"

    if progInfo.toporchild == 'child':
        if not silent:
            print('child window in the foreground, first close, then proceed')
        sendkeys("{escape}")
        time.sleep(0.5)
        foregroundProgInfo = getProgInfo()
        if progInfo.prog != foregroundProgInfo.prog:
            mess = f'after esc key, there is now another program in the foreground: {progInfo.prog}'
            return mess
    
    ## close the window:
    sendkeys(key_close)
    newProgInfo = getProgInfo()
    if newProgInfo.prog != progInfo.prog:
        # OK!
        return True
    
    ## no success, child window must be closed:
    if newProgInfo.toporchild == 'top':
        mess = f'if file needs saving, a child window should be on top now, not {newProgInfo}'
        return mess
    
    ## try to send the no save close dialog key now:
    sendkeys(key_close_dialog)

    progInfo = getProgInfo()
    if progInfo.toporchild == 'child':
        mess = f'killWindow, failed to close child dialog\n\t{progInfo}'
        return mess
    
    if hndle == progInfo.hndle:
        mess = f'killing window failed for {hndle}\n\t{progInfo}'
        return mess
    return True


## initialise ahkexe and ahkscriptfolder:
# GetAhkExe() # is done in next line:
if ahkscriptfolder is None:
    GetAhkScriptFolder()

def test():
    # pylint: disable=C0115, C0116, C0415
    import doctest
    doctest.testmod()


if __name__ ==  "__main__":
    test()
    print(f'ahk_is_active: {ahk_is_active()}')
    Result = getProgInfo()
    print(f'\nresult of getProgModInfo: (start)\n{repr(Result)}')
    Result = ahkBringup("notepad")
    print(f'\nresult of ahkBringup("notepad"):\n{repr(Result)}')
    if Result.hndle:
        killWindow(Result.hndle)
    
    





#
# Python Macro Language for Dragon NaturallySpeaking
#   (c) Copyright 1999 by Joel Gould
#   Portions (c) Copyright 1999 by Dragon Systems, Inc.
#
#pylint:disable=E0611, E0401, R0911, R0912, W0702, C0116

"""natlinkcorefunctions.py

 Quintijn Hoogenboom, January 2008/September 2015

These functions are used by natlinkstatus.py,
and can also used by all the modules,
as the core folder will be in the python path
when natlink is running.

The first function is also, hopefully identical, in
natlinkconfigfunctions, in the configurenatlinkvocolaunimacro folder

getBaseFolder: returns the folder from the calling module
fatalError: raises error again, if new_raise is set, otherwise continues executing
getExtendedEnv(env): gets from os.environ, or from window system calls (CSIDL_...) the
                     environment. Take PERSONAL for HOME and ~
getAllFolderEnvironmentVariables: get a dict of all possible HOME and CSIDL variables,
           that result in a valid folder path
substituteEnvVariableAtStart: substitute back into a file/folder path an environment variable

Note: for extension with %NATLINK% etc. see natlinkstatus.py
    (call getAllEnv, this one first takes Natlink variables and then these extended env variables)

"""
import os
import re
from win32com.shell import shell, shellcon
from natlinkcore import inivars
import natlinkcore   ## for __init__ and getNatlinkUserdirectory

# import win32api
# for extended environment variables:
reEnv = re.compile('(%[A-Z_]+%)', re.I)

# the Natlink Base directory, core/..:
thisBaseFolder = natlinkcore.getThisDir(__file__)
coreDirectory = natlinkcore.getNatlinkDirectory()

# report function:
def fatal_error(message, new_raise=None):
    #pylint:disable=E0704
    """prints a fatal error when running this module"""
    print('natlinkconfigfunctions fails because of fatal error:')
    print()
    print(message)
    print()
    print('Exit Dragon and run the configurenatlink program (via start_configurenatlink.py)')
    print() 
    if new_raise:
        raise new_raise
    raise

# keep track of found env variables, fill, if you wish, with
# getAllFolderEnvironmentVariables.
# substitute back with substituteEnvVariableAtStart.
# and substite forward with expandEnvVariableAtStart
# in all cases a private envDict can be user, or the global dict recentEnv
#
# to collect all env variables, call getAllFolderEnvironmentVariables, see below
recentEnv = {}

def addToRecentEnv(name, value):
    """to be filled for NATLINK variables from natlinkstatus
    """
    recentEnv[name] = value

def deleteFromRecentEnv(name):
    """to possibly delete from recentEnv, from natlinkstatus
    """
    if name in recentEnv:
        del recentEnv[name]

def getFolderFromLibraryName(fName):
    """from windows library names extract the real folder
    
    the CabinetWClass and #32770 controls return "Documents", "Dropbox", "Desktop" etc.
    try to resolve these throug the extended env variables.
    """
    if fName.startswith("Docum"):  # Documents in Dutch and English
        return getExtendedEnv("PERSONAL")
    if fName in ["Muziek", "Music"]:
        return getExtendedEnv("MYMUSIC")
    if fName in ["Pictures", "Afbeeldingen"]:
        return getExtendedEnv("MYPICTURES")
    if fName in ["Videos", "Video's"]:
        return getExtendedEnv("MYVIDEO")
    if fName in ['OneDrive']:
        return getExtendedEnv("OneDrive")
    if fName in ['Desktop', "Bureaublad"]:
        return getExtendedEnv("DESKTOP")
    if fName in ['Quick access', 'Snelle toegang']:
        templatesfolder = getExtendedEnv('TEMPLATES')
        if os.path.isdir(templatesfolder):
            QuickAccess = os.path.normpath(os.path.join(templatesfolder, "..", "Libraries"))
            if os.path.isdir(QuickAccess):
                return QuickAccess
    if fName == 'Dropbox':
        return getDropboxFolder()
    if fName in ['Download', 'Downloads']:
        personal = getExtendedEnv('PERSONAL')
        userDir = os.path.normpath(os.path.join(personal, '..'))
        if os.path.isdir(userDir):
            tryDir = os.path.normpath(os.path.join(userDir, fName))
            if os.path.isdir(tryDir):
                return tryDir
    usersHome = os.path.normpath(os.path.join(r"C:\Users", fName))
    if os.path.isdir(usersHome):
        return usersHome
    if fName in ["This PC", "Deze pc"]:
        return "\\"
    print('cannot find folder for Library name: %s'% fName)
    return ""


def getDropboxFolder(containsFolder=None):
    """get the dropbox folder, or the subfolder which is specified.
    
    Searching is done in all 'C:\\Users' folders, and in the root of "C:"
    (See DirsToTry)
    
    raises OSError if more folders are found (should not happen, I think)
    if containsFolder is not passed, the dropbox main folder is returned
    if containsFolder is passed, this folder is returned if it is found in the dropbox folder
    
    otherwise None is returned.
    """
    results = []
    root = 'C:\\Users'
    dirsToTry = [ os.path.join(root, s) for s in os.listdir(root) if os.path.isdir(os.path.join(root,s)) ]
    dirsToTry.append('C:\\')
    for root in dirsToTry:
        if not os.path.isdir(root):
            continue
        try:
            subs = os.listdir(root)
        except WindowsError:
            continue
        if 'Dropbox' in subs:
            subAbs = os.path.join(root, 'Dropbox')
            subsub = os.listdir(subAbs)
            if not ('.dropbox' in subsub and os.path.isfile(os.path.join(subAbs,'.dropbox'))):
                continue
            if containsFolder:
                result = matchesStart(subsub, containsFolder, caseSensitive=False)
                if result:
                    results.append(os.path.join(subAbs,result))
            else:
                results.append(subAbs)
    if not results:
        return ''
    if len(results) > 1:
        raise OSError('getDropboxFolder, more dropbox folders found: %s')
    return results[0]                 

def matchesStart(listOfDirs, checkDir, caseSensitive):
    """return result from list if checkDir matches, mostly case insensitive
    """
    if not caseSensitive:
        checkDir = checkDir.lower()
    for l in listOfDirs:
        if not caseSensitive:
            ll = l.lower()
        else:
            ll = l
        if ll.startswith(checkDir):
            return l
    return False   
        
            

def getExtendedEnv(var, envDict=None, displayMessage=1):
    """get from environ or windows CSLID

    HOME is environ['HOME'] or CSLID_PERSONAL
    ~ is HOME

    DROPBOX added via getDropboxFolder is this module (QH, December 1, 2018)

    As envDict for recent results either a private (passed in) dict is taken, or
    the global recentEnv.

    This is merely for "caching results"

    """
    if envDict is None:
        myEnvDict = recentEnv
    else:
        myEnvDict = envDict
##    var = var.strip()
    var = var.strip("% ")
    var = var.upper()
    
    if var == "~":
        var = 'HOME'

    if var in myEnvDict:
        return myEnvDict[var]

    if var in os.environ:
        myEnvDict[var] = os.environ[var]
        return myEnvDict[var]

    if var == 'DROPBOX':
        result = getDropboxFolder()
        if result:
            return result
        raise ValueError('getExtendedEnv, cannot find path for "DROPBOX"')

    if var == 'NOTEPAD':
        windowsDir = getExtendedEnv("WINDOWS")
        notepadPath = os.path.join(windowsDir, 'notepad.exe')
        if os.path.isfile(notepadPath):
            return notepadPath
        raise ValueError('getExtendedEnv, cannot find path for "NOTEPAD"')

    # try to get from CSIDL system call:
    if var == 'HOME':
        var2 = 'PERSONAL'
    else:
        var2 = var
        
    try:
        CSIDL_variable =  'CSIDL_%s'% var2
        shellnumber = getattr(shellcon,CSIDL_variable, -1)
    except:
        print('getExtendedEnv, cannot find in environ or CSIDL: "%s"'% var2)
        return ''
    if shellnumber < 0:
        # on some systems have SYSTEMROOT instead of SYSTEM:
        if var == 'SYSTEM':
            return getExtendedEnv('SYSTEMROOT', envDict=envDict)
        return ''
        # raise ValueError('getExtendedEnv, cannot find in environ or CSIDL: "%s"'% var2)
    try:
        result = shell.SHGetFolderPath (0, shellnumber, 0, 0)
    except:
        if displayMessage:
            print('getExtendedEnv, cannot find in environ or CSIDL: "%s"'% var2)
        return ''

    
    result = str(result)
    result = os.path.normpath(result)
    myEnvDict[var] = result
    # on some systems apparently:
    if var == 'SYSTEMROOT':

        myEnvDict['SYSTEM'] = result
    return result

def clearRecentEnv():
    """for testing, clears above global dictionary
    """
    recentEnv.clear()

def getAllFolderEnvironmentVariables(fillRecentEnv=None):
    """return, as a dict, all the environ AND all CSLID variables that result into a folder
    
    Now also implemented:  Also include NATLINK, UNIMACRO, VOICECODE, DRAGONFLY, VOCOLAUSERDIR, UNIMACROUSERDIR
    This is done by calling from natlinkstatus, see there and example in natlinkmain.

    Optionally put them in recentEnv, if you specify fillRecentEnv to 1 (True)

    """
    #pylint:disable=W0603
    global recentEnv
    D = {}

    for k in dir(shellcon):
        if k.startswith("CSIDL_"):
            kStripped = k[6:]
            try:
                v = getExtendedEnv(kStripped, displayMessage=None)
            except ValueError:
                continue
            if len(v) > 2 and os.path.isdir(v):
                D[kStripped] = v
            elif v == '.':
                D[kStripped] = os.getcwd
    # os.environ overrules CSIDL:
    for k in os.environ:
        v = os.environ[k]
        if os.path.isdir(v):
            v = os.path.normpath(v)
            if k in D and D[k] != v:
                print('warning, CSIDL also exists for key: %s, take os.environ value: %s'% (k, v))
            D[k] = v
    if isinstance(fillRecentEnv, dict):
        recentEnv.update(D)
    return D

#def setInRecentEnv(key, value):
#    if key in recentEnv:
#        if recentEnv[key] == value:
#            print 'already set (the same): %s, %s'% (key, value)
#        else:
#            print 'already set (but different): %s, %s'% (key, value)
#        return
#    print 'setting in recentEnv: %s to %s'% (key, value)
#    recentEnv[key] = value

def substituteEnvVariableAtStart(filepath, envDict=None): 
    r"""try to substitute back one of the (preused) environment variables back

    into the start of a filename

    if ~ (HOME) is D:\My documents,
    the path "D:\My documents\folder\file.txt" should return "~\folder\file.txt"

    pass in a dict of possible environment variables, which can be taken from recent calls, or
    from  envDict = getAllFolderEnvironmentVariables().

    Alternatively you can call getAllFolderEnvironmentVariables once, and use the recentEnv
    of this module! getAllFolderEnvironmentVariables(fillRecentEnv)

    If you do not pass such a dict, recentEnv is taken, but this recentEnv holds only what has been
    asked for in the session, so no complete list!

    """
    if envDict is None:
        envDict = recentEnv
    Keys = list(envDict.keys())
    # sort, longest result first, shortest keyname second:
    decorated = [(-len(envDict[k]), len(k), k) for k in Keys]
    decorated.sort()
    Keys = [k for (dummy1,dummy2, k) in decorated]
    for k in Keys:
        val = envDict[k]
        if filepath.lower().startswith(val.lower()):
            if k in ("HOME", "PERSONAL"):
                k = "~"
            else:
                k = "%" + k + "%"
            filepart = filepath[len(val):]
            filepart = filepart.strip('/\\ ')
            return os.path.join(k, filepart)
    # no hit, return original:
    return filepath
       
def expandEnvVariableAtStart(filepath, envDict=None): 
    """try to substitute environment variable into a path name

    """
    filepath = filepath.strip()

    if filepath.startswith('~'):
        folderpart = getExtendedEnv('~', envDict)
        filepart = filepath[1:]
        filepart = filepart.strip('/\\ ')
        return os.path.normpath(os.path.join(folderpart, filepart))
    if reEnv.match(filepath):
        envVar = reEnv.match(filepath).group(1)
        # get the envVar...
        try:
            folderpart = getExtendedEnv(envVar, envDict)
        except ValueError:
            print('invalid (extended) environment variable: %s'% envVar)
        else:
            # OK, found:
            filepart = filepath[len(envVar)+1:]
            filepart = filepart.strip('/\\ ')
            return os.path.normpath(os.path.join(folderpart, filepart))
    # no match
    return filepath
    
def expandEnvVariables(filepath, envDict=None): 
    """try to substitute environment variable into a path name,

    ~ only at the start,

    %XXX% can be anywhere in the string.

    """
    filepath = filepath.strip()
    
    if filepath.startswith('~'):
        folderpart = getExtendedEnv('~', envDict)
        filepart = filepath[1:]
        filepart = filepart.strip('/\\ ')
        filepath = os.path.normpath(os.path.join(folderpart, filepart))
    
    if reEnv.search(filepath):
        List = reEnv.split(filepath)
        #print 'parts: %s'% List
        List2 = []
        for part in List:
            if not part:
                continue
            if part == "~" or (part.startswith("%") and part.endswith("%")):
                try:
                    folderpart = getExtendedEnv(part, envDict)
                except ValueError:
                    folderpart = part
                List2.append(folderpart)
            else:
                List2.append(part)
        filepath = ''.join(List2)
        return os.path.normpath(filepath)
    # no match
    return filepath

def printAllEnvVariables():
    for k in sorted(recentEnv.keys()):
        print("%s\t%s"% (k, recentEnv[k]))

class InifileSection:
    """simulate a part of the registry through inifiles
    
        basic file is natlinkstatus.ini
        basic section is usersettings
        
        So one instance operates only on one section of one ini file!
        
        Now uses the inivars module of Quintijn instead of GetProfileVal system calls.
        
        Only use for readonly or for one section if you want to rewrite data!!
        
        methods:
        set(key, value): set a key=value entry in the section
        get(key, defaultValue=None): get the associated value with
                 key in the current section.
                 if empty or not there, the defaultValue is returned
                 if value = "0" or "1" the integer value 0 or 1 is returned
        delete(key): deletes from section
        keys(): return a list of keys in the section
        __repr__: give contents of a section
        
    """
    def __init__(self, section, filepath):
        """open a section in a inifile
        
        """
        self.section = section
        self.ini = inivars.IniVars(filepath) # want strings to be returned
          
    def __repr__(self):
        """return contents of sections
        """
        L = ["[%s]"% self.section ]
        for k in self.ini.get(self.section):
            v = self.ini.get(self.section, k)
            L.append("%s=%s"% (k, v))
        return '\n'.join(L)
    
    def __iter__(self):
        for item in self.ini.get(self.section):
            yield item         
            
    def get(self, key, defaultValue=None):
        """get an item from a key
        
        """
        if defaultValue is None:
            defaultValue = ''
        else:
            defaultValue = str(defaultValue)
        value = self.ini.get(self.section, key, defaultValue)

        # value = win32api.GetProfileVal(self.section, key, defaultValue, self.filename)
##        if value:
##            print 'get: %s, %s: %s'% (self.section, key, value)
        if value in ("0", "1"):
            return int(value)
        return value

    def set(self, key, value):
        """set an item for akey
        
        0 or empty deletes automatically
        
        """
##        print 'set: %s, %s: %s'% (self.section, key, value)
        if value in [0, "0"]:
            self.delete(key)
        elif not value:
            self.ini.delete(self.section, key)
        else:
            self.ini.set(self.section, key, value)
            self.ini.write()
            # win32api.WriteProfileVal( self.section, key, str(value), self.filename)
            # checkValue = win32api.GetProfileVal(self.section, key, 'nonsens', self.filename)
            # if not (checkValue == value or \
            #                   value in [0, 1] and checkValue == str(value)):
            #     print 'set failed:  %s, %s: %s, got %s instead'% (self.section, key, value, checkValue)
            
    def delete(self, key):
        """delete an item for a key (really set to "")
        
        """
        self.ini.delete(self.section, key)
        self.ini.write()
        # print 'delete: %s, %s'% (self.section, key)
        # value = win32api.WriteProfileVal( self.section, key, None,
        #                                self.filename)
        # checkValue = win32api.GetProfileVal(self.section, key, 'nonsens', self.filename)
        # if checkValue != 'nonsens':
        #     print 'delete failed:  %s, %s: got %s instead'% (self.section, key, checkValue)

    def keys(self):
        Keys = self.ini.get(self.section)
        # Keys =  win32api.GetProfileSection( self.section, self.filename)
        # Keys = [k.split('=')[0].strip() for k in Keys]
        # #print 'return Keys: %s'% Keys
        return Keys

defaultFilename = "natlinkstatus.ini"
defaultSection = 'usersettings'
class NatlinkstatusInifileSection(InifileSection):
    """subclass with fixed filename and section"""
    #pylint:disable=W0613
    
    def __init__(self, filename=defaultFilename, section=defaultSection):
        """get the default inifile:
        
        In the users home diretory/.natlink  the ini file natlinkstatus.ini 
        with section defaultSection
        """        
        natlink_ini_file = natlinkcore.getNatlinkInifile()
        InifileSection.__init__(self, section=section, filepath=natlink_ini_file)


if __name__ == "__main__":
    print(f'this module is in folder: {thisBaseFolder}')
    Vars = getAllFolderEnvironmentVariables()
    for kk in sorted(Vars):
        print('%s: %s'% (kk, Vars[kk]))
        if not os.path.isdir(Vars[kk]):
            print('----- not a directory: %s (%s)'% (Vars[kk], kk))

    print('testing       expandEnvVariableAtStart')
    print('also see expandEnvVar in natlinkstatus!!')
    for p in ("D:\\natlink\\unimacro", "~/unimacroqh",
              "%HOME%/personal",
              "%WINDOWS%\\folder\\strange testfolder"):
        expanded = expandEnvVariableAtStart(p)
        print('expandEnvVariablesAtStart: %s: %s'% (p, expanded))
    print('testing       expandEnvVariables')  
    for p in ("%DROPBOX%/QuintijnHerold/jachthutten", "D:\\%username%", "%NATLINK%\\unimacro", "%UNIMACROUSER%",
              "%HOME%/personal", "%HOME%", "%personal%"
              "%WINDOWS%\\folder\\strange testfolder"):
        expanded = expandEnvVariables(p)
        print('expandEnvVariables: %s: %s'% (p, expanded))

    # testIniSection = NatlinkstatusInifileSection()
    # print testIniSection.keys()
    # testIniSection.set("test", "een test")
    # testval = testIniSection.get("test")
    # print 'testval: %s'% testval
    # testIniSection.delete("test")
    # testval = testIniSection.get("test")
    # print 'testval: %s'% testval
    print('recentEnv: %s'% len(recentEnv))
    np = getExtendedEnv("NOTEPAD")
    print(np)
    for lName in ['Snelle toegang', 'Quick access', 'Documenten', 'Documents', 'Muziek', 'Afbeeldingen', 'Dropbox', 'OneDrive', 'Desktop', 'Bureaublad']:
        f = getFolderFromLibraryName(lName)
        print('folder from library name %s: %s'% (lName, f))



import win32com.client
import time
import os
import os.path
from pythoncom import com_error
from dtactions.unimacro.actionclasses.actionbases import AllActions
from dtactions.unimacro import unimacroutils as natqh


class ExcelActions(AllActions):
    """attach to excel and perform necessary actions
    """
    appList = [] # here goes, for all instances, the excel application object
    positions = {} # dict with book/sheet/(col, row) (tuple)
    rows = {} # dict with book/sheet/row  (str)
    columns = {} # dict with book/sheet/col (str)
    def __init__(self, progInfo):
        AllActions.__init__(self, progInfo)
        self.connect() # sets self.app
        if not self.app:
            return
        self.update(progInfo)

    def reset(self, progInfo=None):
        """reset progInfo and prev variables. Leave self.app
        """
        AllActions.reset(self, progInfo)
        self.prevBook = self.prevSheet = self.prevPosition = None

    def update(self, progInfo):
        print('update of ExcelActions, %s'% self.topHandle)
        if self.topHandle != progInfo[3]:
            print('reset, topHandle: %s, progInfo: %s'% (self.topHandle, repr(progInfo)))
            self.reset(progInfo)
            if self.app:
                print('wrong app for excel, window handle invalid, app is connected to %s, forground window has %s\nPlease close conflicting Excel instance!'% (self.topHandle, progInfo[3]))
                self.disconnect() # sets self.app to None
            else:
                print('try to reconnect to %s'% self.topHandle)
                self.connect()
        if self.app:
            self.checkForChanges(progInfo)
        else:
            print('no valid excel instance available for hndle: %s'% self.topHandle)
        
    def isInForeground(self, app=None, title=None, progInfo=None):
        """return True if app is in foreground
        """
        thisApp = app or self.app
        title = title or self.topTitle
        title = title.lower()
        
        if not thisApp:
            return False
        if thisApp.Workbooks.Count == 0:
            print('Got excel app with no workbooks')
            return False
        try:
            name = thisApp.ActiveWorkbook.Name
        except AttributeError:
            print('workbook is not saved yet.')
            return True
        #print 'activeworkbook: %s'% name
        if title.find(name.lower()) >= 0:
            return True
        print('============\nexcel, isInForeground: cannot find name \n"%s" \nin title: \n"%s"\nProbably you have more excel instances open. Unimacro only can connect to this one instance.\n'% (name, title))
        return False        
        
    def checkForChanges(self, progInfo=None):
        """return 1 if book, sheet or position has changed since previous call
        """
        progInfo = progInfo or unimacroutils.getProgInfo()
        changed = 0
        if not self.app:
            self.progInfo = progInfo
            self.connect()
            if not self.app:
                return
        
        try:
            visible = self.app.Visible
        except:
            self.app = None
            self.connect()
            if not self.app:
                return
        
        title = self.topTitle = progInfo[1]

        if not self.isInForeground(app=self.app, title=title, progInfo=progInfo):
            print('not in foreground, excel')
            if self.prevBook:
                changed = 8 # from foreground into background
                self.prevBook = self.prevSheet = self.prevPosition = None
            self.book = self.sheet = self.Position = None
            self.currentRow = self.currentLine = None
            self.currentColumn = None
            self.progInfo = progInfo
            return


        self.book = self.app.ActiveWorkbook
        
        if self.book is None:
            print('ExcelAction, no active workbook')
            self.bookName = ''
            self.sheetName = ''
            self.sheet = None
        else:
            self.bookName = self.book.Name
            self.sheet = self.app.ActiveSheet
            self.sheetName = self.sheet.Name
        if self.prevBook != self.bookName:
            self.prevBook = self.bookName
            changed += 4
        if self.prevSheet != self.sheetName:
            self.prevSheet = self.sheetName
            changed += 2
        if changed:
            # print 'excel-actions: update book and/or sheet variables'
            self.positions.setdefault(self.bookName,{})
            self.positions[self.bookName].setdefault(self.sheetName, [])
            self.columns.setdefault(self.bookName,{})
            self.columns[self.bookName].setdefault(self.sheetName, [])
            self.rows.setdefault(self.bookName,{})
            self.rows[self.bookName].setdefault(self.sheetName, [])
            self.currentPositions = self.positions[self.bookName][self.sheetName]
            self.currentColumns = self.columns[self.bookName][self.sheetName]
            self.currentRows = self.rows[self.bookName][self.sheetName]
        if self.sheet:
            #print 'excel-actions: %s, update current position'% self.sheet

            cr = self.savePosition()
            self.currentPosition = cr
            self.currentRow = self.currentLine = cr[1]
            self.currentColumn = cr[0]
            #print 'currentColumn: %s, currentRow: %s, currentLine: %s'% (self.currentColumn, self.currentRow, self.currentLine)
            if cr != self.prevPosition:
                self.prevPosition = cr
                changed += 1
            #else:
            #    print 'same position: %s'% repr(cr)
        else:
            print('excel-actions, no self.sheet.')
        #print 'return code checkForChanges: %s'% changed
        return changed
    
    #connect to programs:
    def connect(self):
        """connect to excel set self.app
        
        set self.app to None if no excel instance available
        also reset the relevant variables...
        return value irrelevant
        """
        print('Connecting to to excel...')
        self.prevBook = self.prevSheet = self.prevPosition = None
        if self.prog != 'excel':
            print("excel-actions, should only be called when excel is the foreground window, not: %s"% self.prog)
            self.app = None
            return
        title = self.topTitle
        try:
            self.app = win32com.client.GetActiveObject(Class="Excel.Application")
        except com_error:
            print('Cannot attach to Excel.')
            self.app = None
            return
        if self.isInForeground():
            return  # OK
        else:
            print('ExcelActions: cannot attach to Excel, because instance is not in the foreground.')
            print('Probably there are more excel instances active.')
            print('Please close excel instances and try again')
            self.disconnect()
            return


    def recentMatchesTitle(self, recentFile, windowTitle):
        """check the recent file with the current window title
        """
        p, name = os.path.split(recentFile)
        if windowTitle.find(name) >= 0:
            return True
        

    def disconnect(self):
        """disconnect from excel and set self.app to None
        
        reset the relevant variables
        by removing the connection (not powerful enough)
        """
        # if self.app:
            # try:
            #     self.app.Quit()
            # except com_error:
            #     pass
        self.reset()
        self.app = None

    def savePosition(self):
        """save current position in positions dict
        
        returns the current (col, row) tuple
        """
        cr = self.getCurrentPosition()
        c, r = cr
        self.pushToListIfDifferent(self.currentPositions, cr)
        self.pushToListIfDifferent(self.currentRows, r)
        self.pushToListIfDifferent(self.currentColumns, c)
        
        return cr
    
    def getBookFromTitle(self, title, app=None):
        """return the book name corresponding with title (being the window title)
        """
        bookList = self.getBooksList(app=app)
        if bookList:
            for b in bookList:
                if title.find(b) >= 0:
                    print('found book in instance: %s'% b)
                    return b
        print('not found title in instance, bookList: %s'% bookList)

    def getBookNameFromTitle(self, title):
        """return the book name corresponding with title (being the window title)
        """
        excelTexts = ['Microsoft Excel -', '- Microsoft Excel']
        print('getBookNameFromTitle: %s'% title)
        if title.find('['):
            title = title.split('[')[0].strip()
        for e in excelTexts:
            if title.find(e) >= 0:
                title = title.replace(e, '').strip()
        print('result: %s'% title)
        return title
            
        
    
    def getBooksList(self,app=None):
        """get list of strings, the names of the open workbooks
        """
        app = app or self.app
        if app:
            books = []
            for i in range(app.Workbooks.Count):
                b = app.Workbooks(i+1)
                books.append(b.Name)
            return books
        return []

    def getCurrentLineNumber(self, handle=None):
        if not self.app: return
        lineStr = self.getCurrentPosition()[1]
        return int(lineStr)
    
    
    def getSheetsList(self, book=None):
        """get list of strings, the names of the open workbooks
        """
        if book is None:
            book = self.book
        elif isinstance(book, str):
            book = self.app.Workbooks[book]
        
        if book:
            sheets = []
            for i in range(book.Sheets.Count):
                s = self.app.Worksheets(i+1)
                sheets.append(s.Name) # assume this is Unicode...
            return sheets

    
    def selectSheet(self, sheet):
        """select the sheet by name of number
        """
        self.app.Sheets(sheet).Activate()
    
    def getCurrentPosition(self):
        """return row and col of activecell
        
        as a side effect remember the (changed position)
        """
        if not (self.app and self.book and self.sheet):
            return None, None
        ac = self.app.ActiveCell
        comingFrom = ac.Address
        #print 'activecell: %s'% comingFrom
        cr = [value.lower() for value in comingFrom.split("$") if value] # assume value unicode
        if len(cr) == 2:
            #print 'currentposition, lencr: %s, cr: %s'% (len(cr), cr)
            return tuple(cr)
        else:
            print('excel-actions, no currentposition, lencr: %s, cr: %s'% (len(cr), cr))
            return None, None

    def pushToListIfDifferent(self, List, value):
        """add to list (in place) of value differs from last value in list
        
        for positions, value is (c,r) tuple
        """
        if not value:
            return
        if List and List[-1] == value:
            return
        List.append(value)
    
    def getPreviousRow(self):
        """return the previous row number
        """
        cr = self.getCurrentPosition()
        c, r = cr
        while 1:
            newr = self.popFromList(self.currentRows)
            if newr is None:
                return
            if r != newr:
                return newr

    def getPreviousColumn(self):
        """return the previous col letter
        """
        cr = self.getCurrentPosition()
        c, r = cr
        while 1:
            newc = self.popFromList(self.currentColumns)
            if newc is None:
                return
            if c != newc:
                return newc
    
    def popFromList(self, List):
        """pop from list a different value than currentValue
        
        and return None if List is exhausted
        """
        if List:
            value = List.pop()
            return value


    # functions that do an action from the action.py module, in case of excel:
    # one parameter should be given    
    def metaaction_gotoline(self, rowNum):
        """overrule for gotoline meta-action
        
        goto line in the current column
        """
        rowNum = str(rowNum)
        if not self.app:  return
        cPrev, rPrev = self.getCurrentPosition()
        if rowNum == rPrev:
            print('row already selected')
            return 1
        try:
            range = cPrev + rowNum
            #print 'current range: %s, %s'% (rPrev, cPrev)
            sheet = self.app.ActiveSheet
            #print 'app: %s, sheet: %s (%s), range: %s'% (app, sheet, sheet.Name, range)
            sheet.Range(range).Select()
            return 1
        except:
            print('something wrong in excel-actions, metaaction_gotoline.')
            return
            
    def metaaction_selectline(self, dummy=None):
        """select the current line
        """
        if not self.app: return
        self.app.ActiveCell.EntireRow.Select()
        return 1

    def metaaction_remove(self, dummy=None):
        """remove the selection, assume rows or columns are selecte
        """
        if not self.app: return
        self.app.Selection.Delete()

        return 1
        
    def metaaction_insert(self, dummy=None):
        """insert the number of lines that are selected
        """
        if not self.app: return
        self.app.Selection.Insert()

        return 1

    metaaction_lineinsert = metaaction_insert

    def metaaction_pasteinsert(self, dummy=None):
        """insert the number of lines that are selected
        """
        if not self.app: return
        self.app.Selection.Insert()
        self.app.CutCopyMode = False

        return 1

    def metaaction_selectpreviousline(self, dummy=None):
        """select the previous line with respect to the activecell
        """
        if not self.app: return
        wantedLine = int(self.currentRow) - 1
        self.metaaction_gotoline(wantedLine)
        self.app.ActiveCell.EntireRow.Select()
        return 1

    def metaaction_movetotopofselection(self, dummy=None):
        """select first line of a selected range
        """
        if not self.app: return
        self.app.Selection.Rows(1).Select()
        return 1

    def metaaction_movetobottomofselection(self, dummy=None):
        """select first line of a selected range
        """
        if not self.app: return
        nRows = self.app.Selection.Rows.Count
        self.app.Selection.Rows(nRows).Select()
        return 1
    def metaaction_movecopypaste(self, dummy=None):
        """insert the clipboard after a movecopy action.
        """
        if not self.app: return
        self.app.ActiveCell.EntireRow.Insert()
        self.app.CutCopyMode = False
        return 1
        
    def metaaction_lineback(self, dummy=None):
        """goes back to the previous row
        """
        if not self.app: return
        #self.app.ActiveCell.EntireRow.Select()
        prevRow = self.getPreviousRow()
        print('prevRow: %s'% prevRow)
        if prevRow:                
            self.gotoRow(prevRow)
        return 1
        
if __name__ == '__main__':
    progInfo = ('excel', 'Microsoft Excel - Map1', 'top', 921280)
    excel = ExcelActions(progInfo)
    if excel.app:
        #excel.app.Visible = True
        print('activeCell: %s'% excel.app.ActiveCell)
        print('books: %s'% excel.app.Workbooks.Count)
        print('foreground: %s'%excel.isInForeground())
        print('now click on excel please')
        print(excel.app.hndle)
        excel.metaaction_gotoline(345)
        time.sleep(5)
    else:
        print('no excel.app: %s'% excel.app)
            



#
# This file was part of Dragonfly.
# (c) Copyright 2007, 2008 by Christo Butcher
# Licensed under the LGPL.
#
# It is now improved and augmented for the Natlink project
# by Quintijn Hoogenboom for wider use 15/7/2019, ..., March 2021
# See also unittestClipboard.py in Unimacro/unimacro_test

"""
This file implements an interface to the Windows system clipboard.
"""
#pylint:disable=C0116, C0321, R1710, W0702, R0913, R0912, W0622
import copy
import time
import win32clipboard
import win32con
from dtactions.sendkeys import sendkeys
#===========================================================================

class Clipboard:
    """Clipboard class, manages getting and setting the windows clipboard
    """

    #-----------------------------------------------------------------------
    format_text      = win32con.CF_TEXT
    
    format_oemtext   = win32con.CF_OEMTEXT
    format_unicode   = win32con.CF_UNICODETEXT
    format_locale    = win32con.CF_LOCALE
    format_hdrop     = win32con.CF_HDROP
    format_names = {
                    format_text:     "text",
                    format_oemtext:  "oemtext",
                    format_unicode:  "unicode",
                    format_locale:   "locale",
                    format_hdrop:    "hdrop",
                   }

    @classmethod
    def get_system_text(cls):
        """get text from the clipboard
        
        from natlinkclipboard import Clipboard
        
        simply call: text = Clipboard.get_system_text()
        
        as alias also Get_text can be used, so: text = Clipboard.Get_text()
        """
        if not OpenClipboardCautious():
            print('Clipboard, get_system_text: could not open clipboard')
            return
        try:
            for _format in cls.format_unicode, cls.format_text, cls.format_oemtext:
                try:
                    content = win32clipboard.GetClipboardData(_format)
                    if content: break
                except:
                    continue
            else:
                print('Clipboard, get_system_text, no content found')
                content = ""
            if content:
                content = content.replace('\0', '')
                if content.find('\r\n') >= 0:
                    content = content.replace('\r\n', '\n')
            return content
        finally:
            win32clipboard.CloseClipboard()

    Get_text = get_system_text

    @classmethod
    def set_system_text(cls, content):
        """set text to the clipboard
        
        First the clipboard is emptied. 
        
        This method fails when not in elevated mode.
        
        As alias, you can also call: Clipboard.Set_text("abacadabra")
        """
        print(f'set to clipboard: {content}')
        # content = str(content)
        if not OpenClipboardCautious():
            print('Clipboard, set_system_text: could not open clipboard')
            return
        clipNum = win32clipboard.GetClipboardSequenceNumber()
        # print 'clipboard number: %s'% clipNum
        try:
            win32clipboard.EmptyClipboard()
            if isinstance(content, bytes):
                _format = cls.format_text
            elif isinstance(content, str):
                _format = cls.format_unicode
            
            win32clipboard.SetClipboardData(_format, content)
        except:
            print(f'Clipboard, cannot set text to clipboard: {content}')
        finally:
            clipNum2 = win32clipboard.GetClipboardSequenceNumber()
            if clipNum2 == clipNum:
                print(f'Clipboard, did not increment clipboard number: {clipNum}')
            win32clipboard.CloseClipboard()
    Set_text = set_system_text

    @classmethod
    def Get_clipboard_formats(cls):
        """returns a list of format types of current clipboard
        
        This is mainly meant for debugging purposes.
        """
        if not OpenClipboardCautious():
            print('get_clipboard_formats, could not open clipboard')
            return
        try:
            # 
            formats = _get_clipboard_formats_open_clipboard()
            return formats
        finally:
            win32clipboard.CloseClipboard()

    @classmethod
    def get_system_folderinfo(cls, waiting_time=0.05):
        """returns a tuple of file/folder info of selected files or folders
        
        As alias use Get_folderinfo, Get_hdrop or get_system_hdrop
        
        win32con.CF_HDROP is the parameter for calling this type of clipboard data
        
        """
        if not OpenClipboardCautious(waiting_time=waiting_time):
            print('Clipboard, get_system_folderinfo, could not open clipboard')
            return
        try:
            for i in range(3):
                try:
                    folderinfo = win32clipboard.GetClipboardData(cls.format_hdrop)
                    break
                except:
                    time.sleep(0.1)                    
                    continue
                else:
                    print(f'{i} folderinfo: {folderinfo}')
                    if folderinfo:
                        print('got it!')
                        return folderinfo
            else:
                print('Clipboard, get_system_folderinfo: no folder info found.')
                return
            return folderinfo
        finally:
            win32clipboard.CloseClipboard()
    
    get_system_hdrop = get_system_folderinfo
    Get_folderinfo = get_system_folderinfo
    Get_hdrop = get_system_folderinfo

    #-----------------------------------------------------------------------
    def __init__(self, contents=None, text=None, save_clear=False, debug=None, waiting_interval=None, waiting_iterations=None):
        """initialisation of clipboard instance.
        
        save_clear can be set to True, current clipboard contents is saved and cleared
               saved contents are kept in self._backup and will be retrieved
               when instance is destroyed.
        contents: can be set as initial contents of the clipboard (not tested, 2019)
        text: can be set as initial text contents of the clipboard (not tested, 2019)
        debug: default off, 1: important messages are printed, > 1 more messages are printed
        
        """
        self._contents = {}
        self._backup = None
        self.debug = debug or 0
        self.waiting_interval = waiting_interval or 0.025
        self.waiting_iterations = waiting_iterations or 10
        if not OpenClipboardCautious():
            if self.debug: print('Warning Clipboard: at initialisation could not open the clipboard')
            return
        self.current_sequence_number = win32clipboard.GetClipboardSequenceNumber()
        if self.debug > 1: print(f'current_sequence_number: {self.current_sequence_number}')

        # If requested, retrieve current system clipboard contents.
        if save_clear:
            self.copy_from_system(save_clear=save_clear)

        # Process given contents for this Clipboard instance.
        if contents:
            try:
                self._contents = dict(contents)
            except Exception as exc:
                raise TypeError(f'Clipboard: Invalid contents: "{contents}"') from exc

        # Handle special case of text content.
        if not text is None:
            self._contents[self.format_unicode] = str(text)

    def __del__(self):
        """restore clipboard if self._backup contains data
        
        This is so if save_clear is True when the instance was created,
        and when the clipboard held data at that moment.
        
        Note: need elevated mode for setting the clipboard...
        """
        if self._backup:
            self.restore()

    def __str__(self):
        arguments = []
        skip = []
        if not self._contents:
            arguments = ["(empty)"]
        else:
            text = self.get_text()
            if text:
                if len(text) > 20:
                    shorter = text[:10] + ' // ' + text[-10:]
                    arguments.append(repr(shorter))
                else:
                    arguments.append(repr(text))
                    
                skip.append(self.format_text)
                skip.append(self.format_unicode)
                skip.append(self.format_oemtext)
            else:
                arguments.append("(no_text)")
            for _format in sorted(self._contents.keys()):
                if _format in skip:
                    continue
                if _format in self.format_names:
                    arguments.append(self.format_names[_format])
                else:
                    arguments.append(repr(_format))
        arguments = ", ".join(str(a) for a in arguments)
        return f'{self.__class__.__name__}({arguments})'

    def copy_from_system(self, formats=None, save_clear=False):
        """Copy the Windows system clipboard contents into this instance.

            Arguments:
             - *formats* (iterable, default: None) -- if not None, only the
               given content formats will be retrieved.  If None, all
               available formats will be retrieved.
             - *save_clear* (boolean, default: False) -- if true, the Windows
               system clipboard will be saved in self._backup, and
               cleared after its contents have been retrieved.
               Will be restored from self._backup when the instance is destroyed.
               If false contents are retrieved in self._contents
        """
        result = self.wait_for_clipboard_change()
        if result is None:
            print("no clipboard change")
            return
        if not OpenClipboardCautious():
            if self.debug: print('Clipboard copy_from_system, could not open clipboard')
            return

        try:                
        # Determine which formats to retrieve.
            contents = self._get_clipboard_data_from_system(formats=formats)

            # Retrieve Windows system clipboard content.
            if save_clear:
                if contents:
                    self._backup = copy.copy(contents)
                    if self.debug > 1: print(f'Clipboard, set backup to: {repr(self._backup)}')
                else:
                    self._backup = None
                self._contents = None
                win32clipboard.EmptyClipboard()
            else:
                if contents:
                    self._contents = copy.copy(contents)
                else:
                    self._contents = None
            return contents
        finally:    
            # Clear the system clipboard, if requested, and close it.
            self.current_sequence_number = win32clipboard.GetClipboardSequenceNumber()
            win32clipboard.CloseClipboard()

    def _get_clipboard_data_from_system(self, formats=None):
        """once the clipboard is opened, just get the clipboard data
        
        meant as internal functions, called from copy_from_system
        return the contents as a dict, see for example method get_text
        also leave the contents in self._contents
        """
        contents = {}
        self._contents = {}
        if not formats:
            formats = _get_clipboard_formats_open_clipboard()
        elif isinstance(formats, int):
            formats = (formats,)

        # Verify that the given formats are valid.
        if not formats:
            if self.debug > 1: print('_get_clipboard_data_from_system, no formats available, empty clipboard...')
            return contents

        if formats:
            for _format in formats:
                try:
                    content = win32clipboard.GetClipboardData(_format)
                    contents[_format] = content
                except:
                    pass  # unknown formats, which cannot be handled
            self._contents = contents
            return contents

    def copy_to_system(self, data=None, clear=True):
        """Copy the contents of this instance to the Windows clipboard

            Arguments:
            - data: text or dict of clipboard items (format, content) pairs
             - *clear* (boolean, default: True) -- if true, the Windows
               system clipboard will be cleared before this instance's
               contents are transferred.

        """
        if not OpenClipboardCautious():
            print('copy_to_system, could not open clipboard')
            return
        try:
            # Clear the system clipboard, if requested.
            if clear:
                try:
                    win32clipboard.EmptyClipboard()
                except:
                    if self.debug > 1: print('Clipboard, cannot EmptyClipboard, need more rights, can also not restore backup of clipboard')
    
            # Transfer content to Windows system clipboard.
            data = data or self._contents
            if data is None:
                return
            if isinstance(data, str):
                self.get_text(data)
            elif isinstance(data, dict):
                for _format, content in list(data.items()):
                    win32clipboard.SetClipboardData(_format, content)
            else:
                
                if self.debug: print(f'Clipboard, copy_to_system, invalid type of data: {type(data)}')
                if self.debug > 1: print(f'data: {repr(data)}\n========')
                return 
        finally:
            self.current_sequence_number = win32clipboard.GetClipboardSequenceNumber()
            win32clipboard.CloseClipboard()

    def Copy_and_get_clipboard(self, keyscopy=None):
        """send the keystrokes for copy, and collect the clipboard
        """
        self.save_sequence_number()
        keyscopy = keyscopy or '{ctrl+c}'
        sendkeys(keyscopy)
        content = self.get_text()
        return content
        
    
    def Set_text_and_paste(self, t):
        """a one shot function to past text back into the application
        """
        if isinstance(t, str) and t:
            self.copy_to_system(data=t)
        #### to be finished


    def restore(self):
        """restore the _backup to the system clipboard
        """
        if self._backup:
            self.copy_to_system(self._backup, clear=True)
        else:
            if self.debug: print('Clipboard restore, nothing to restore')
            
    def clear_clipboard(self):
        """Empty the clipboard and clear the internal clipboard data
        
        assume the clipboard is open
        
        this will be done at init phase with save_clear == True
        """
        if self.debug > 1:
            print(f'clear_clipboard, before start, sequence number: {self.current_sequence_number}')

        if not OpenClipboardCautious():
            print('copy_to_system, could not open clipboard')
            return
        try:        
            win32clipboard.EmptyClipboard()
            self._contents = None
        except:
            print("clipboard, clear_clipboard, could not clear the clipboard")
        finally:
            self.current_sequence_number = win32clipboard.GetClipboardSequenceNumber()
            if self.debug > 1:
                print(f'clear_clipboard, new sequence number: {self.current_sequence_number}')
            win32clipboard.CloseClipboard()
    
    def has_format(self, format):
        """Determine whether this instance has content for the given format

            Arguments:
             - *format* (int) -- the clipboard format to look for.
        """
        return format in self._contents

    def get_format(self, format):
        """Retrieved this instance's content for the given *format*.

            Arguments:
             - *format* (int) -- the clipboard format to retrieve.

            If the given *format* is not available, a *ValueError*
            is raised.
        """
        try:
            return self._contents[format]
        except KeyError as exc:
            raise ValueError(f'Clipboard format not available: {format}') from exc

    def set_format(self, format, content):
        self._contents[format] = content

    def has_text(self):
        """ Determine whether this instance has text content. """
        if self._contents:
            return (self.format_unicode in self._contents
                    or self.format_text in self._contents)
        return False

    def get_text(self, replaceNullChar=True):
        """get the text (mostly unicode) contents of the clipboard
        
        This method first does a copy from system.
        
        If no text content available, return ""
        """
        contents = self.copy_from_system(formats = [self.format_unicode, self.format_text])
        text = ""
        if contents:
            if self.format_unicode in contents:
                text = contents[self.format_unicode]
                if text.find('\r\n') >= 0:
                    text = text.replace('\r\n', '\n')
                if text.find('\0') >= 0:
                    if replaceNullChar:
                        text = text.replace('\0', '')
            elif self.format_text in contents:
                text = contents[self.format_text]
            else:
                print(f'get contents, but no expected format: {contents.keys()}')
                text = ""
        if text.find('\r\n') >= 0:
            text = text.replace('\r\n', '\n')
        return text

    def set_text(self, content):
        self._contents[self.format_unicode] = str(content)

    text    = property(
                       lambda self:    self.get_text(),
                       lambda self, d: self.set_text(d)
                      )

    def get_folderinfo(self):
        """Retrieve this instance's folderinfo (also hdrop)
        
        do a copy_from_system automatically
           
        This should be a tuple of valid paths. The paths are not checked.

        If no valid info, return None
        """
        self.copy_from_system()
        if self.format_hdrop in self._contents:
            result = self._contents[self.format_hdrop]
            if isinstance(result, tuple):
                return result

    def save_sequence_number(self):
        """get the Clipboard Sequence Number and store in instance
        
        It is set in self.current_sequence_number, no return
        """
        self.current_sequence_number = win32clipboard.GetClipboardSequenceNumber()
        
    def wait_for_clipboard_change(self):
        """wait a few steps until the clipboard is not changed.
        
        The previous Clipboard Sequence Number should be in
        the instance variable self.current_sequence_number
        
        This value is set when opening the clipboard, or in this internal function.
        
        When in doubt, call _set_sequence_number, before doing a copy!
        
        return True if changed
        """
        try:
            if self.waiting_iterations <= 0:
                print('Clipboard, _wait_for_clipboard_change, no valid value waiting_iterations {self.waiting_iterations}')
                return
        except ValueError:
            print('Clipboard, _wait_for_clipboard_change, no valid value for waiting_iterations {self.waiting_iterations}')
            return
        try:
            if self.waiting_interval <= 0:
                print('Clipboard, _wait_for_clipboard_change, no valid value waiting_interval {self.waiting_interval}')
                return
        except ValueError:
            print('Clipboard, _wait_for_clipboard_change, no valid value for waiting_interval {self.waiting_interval}')
            return

        for i in range(self.waiting_iterations):
            new_sequence_number = win32clipboard.GetClipboardSequenceNumber()
            if new_sequence_number > self.current_sequence_number:
                self.current_sequence_number = new_sequence_number
                if i:
                    if self.debug: print(f'---Clipboard changed after {i} steps of {self.waiting_interval:4f}')
                else:
                    if self.debug > 1: print('wait_for_clipboard_change, clipboard changed immediately')
                # time.sleep(w_time)
                return i
            time.sleep(self.waiting_interval)
        # no result:
        time_waited = self.waiting_interval*self.waiting_iterations
        if self.debug: print(f'Clipboard, no change in clipboard in {time_waited:.4f} seconds')
    
    get_hdrop = get_folderinfo

def _get_clipboard_formats_open_clipboard():
    """return clipboard formats available, when clipboard is open
    
    """
    formats = []
    f = win32clipboard.EnumClipboardFormats(0)
    while f:
        formats.append(f)
        f = win32clipboard.EnumClipboardFormats(f)
    return formats

def OpenClipboardCautious(nToTry=4, waiting_time=0.1):
    """sometimes, wait a little before you can open the clipboard...
    """
    for i in range(nToTry):
        try:
            win32clipboard.OpenClipboard()
        except:
            time.sleep(waiting_time)
            continue
        else:
            wait = (i+2)*waiting_time
            if i:
                print(f'had to wait, and extra wait {i} OpenClipboardCautious: {wait:.4f} seconds')
            time.sleep(waiting_time)
            return True
   
